#ifndef _ClientDPDPmain_H_
#define _ClientDPDPmain_H_

/* Partially from http://netbook.cs.purdue.edu*/


#include "DPDPFlexVerifier.h"
#include <boost/asio.hpp>
#include <boost/bind.hpp>
#include <iostream>
#include <vector>
#include "connection.hpp" // Must come before boost/serialization headers.
#include <boost/serialization/vector.hpp>

#include <stdio.h>
#include <string.h>
#include <boost/serialization/vector.hpp>
#include "GroupRSA.h"
#include "RankBasedSkipList.h"
#include "PacketClient.h"
#include "PacketServer.h"

#define PROTOPORT       5195            /* default protocol port number */
using boost::asio::ip::tcp;

/*------------------------------------------------------------------------
 * Program:   	ClientDPDPmain
 *
 * Purpose:   	connect to the host and port provided, initialize FlexDPDP variables,
 * 				lets Client use the menu where she can send her file, challenge her file and commit changes done
 *
 *
 * Note:      	There are helper classes below where the names of classes are self explanatory.
 *
 // Partially from http://netbook.cs.purdue.edu
 *------------------------------------------------------------------------
 */
// Partially:
// Copyright (c) 2003-2012 Christopher M. Kohlhoff (chris at kohlhoff dot com)
//
// Distributed under the Boost Software License, Version 1.0. (See accompanying
// file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)

namespace s11n_example {
    
    class CommitClass {
        
    public:
        
        CommitClass(boost::asio::io_service& io_service,
                    const std::string& host, const std::string& service, char* fileName, DPDPFlexVerifier* verif, int i = 0);
        void connect(boost::asio::io_service& io_service, const std::string& host, const std::string& service );
        void handle_connect(const boost::system::error_code& e);
        void handle_write(const boost::system::error_code& e, vector<DiffSkipList> ad, int seed/*, connection_ptr conn*/);
        void handle_read(const boost::system::error_code& e, vector<DiffSkipList> diffs, int seed);
        int getID(){ return id; }
        
    private:
        
        vector<int> challenges;
        char *fileName;
        connection connection_;
        DPDPFlexVerifier* verifier;
        PacketServer packet;
        int id;
    };
    
    class TransmissionClass { //Basic class in which the file is divided into blocks are sent to the server
        
    public:
        
        TransmissionClass(boost::asio::io_service& io_service,
                          const std::string& host, const std::string& service, char* fN, DPDPFlexVerifier* verif, int i = 0);
        void connect(boost::asio::io_service& io_service, const std::string& host, const std::string& service );
        void handle_connect(const boost::system::error_code& e);
        /// Handle completion of a write operation.
        void handle_write(const boost::system::error_code& e/*, connection_ptr conn*/);
        void handle_read(const boost::system::error_code& e);
        int getID() { return id; }
        
    private:
        
        char *fileName;
        connection connection_;
        DPDPFlexVerifier* verifier;
        PacketServer packet;
        int id;
    };
    
    class ChallengeClass { //in this class client generates a seed and sends it to the server by just keeping which bytes she challenged in mind. Then she can verify the blocks arrived.
        
    public:
        
        ChallengeClass(boost::asio::io_service& io_service,
                       const std::string& host, const std::string& service, char* fileName, DPDPFlexVerifier* verif, int i = 0, int se = 1, int cCount = 1);
        void connect(boost::asio::io_service& io_service, const std::string& host, const std::string& service );
        void handle_connect(const boost::system::error_code& e);
        /// Handle completion of a write operation.
        void handle_write(const boost::system::error_code& e, vector<int> challenges, int seed/*, connection_ptr conn*/);
        void handle_read(const boost::system::error_code& e, vector<int> challenges, int seed);
        int getID() { return id; }
        
    private:
        
        vector<int> challenges;
        char *fileName;
        connection connection_;
        DPDPFlexVerifier* verifier;
        PacketServer packet;
        int id;
        int sec, clientCount;
    };
    
    /*MAIN CLASS WHERE THE UPPER HELPER CLASSES ARE INITIATED AND USED*/
    class ClientDPDPmain {
        
    public:
        
        ClientDPDPmain(boost::asio::io_service& io_service,
                       const std::string& host, const std::string& service, int i = 0);
        void connect(boost::asio::io_service& io_service, const std::string& host, const std::string& service );
        /// Handle completion of a connect operation.
        void handle_connect(const boost::system::error_code& e);
        /// Handle completion of a read operation.
        void handle_read(const boost::system::error_code& e);
        /// Handle completion of a write operation.
        void handle_write(const boost::system::error_code& e/*, connection_ptr conn*/);
        int printMenu();
        void startOperations(int choice);
        
        int getID(){ return id; }
        void setID(int i) { id = i; }
        
    private:
        
        /*DPDP RELATED FIELDS*/
        DPDPFlexVerifier verifier;
        /*CLIENT SERVER RELATED FIELDS*/
        /// The connection to the server.
        char *fileName;
        string fName;
        connection connection_;
        
        PacketServer packet;
        int id;
        string theHost;
        string theService;
        
    };
    
    
    
    
    
    
    
    
    //START CommitClass:: function definitions
    
    CommitClass::CommitClass(boost::asio::io_service& io_service,
                             const std::string& host, const std::string& service, char* fileName, DPDPFlexVerifier* verif, int i)
    : connection_(io_service)
    {
        id = i;
        verifier = verif;
        connect(io_service, host, service);
    }
    
    void CommitClass::connect(boost::asio::io_service& io_service, const std::string& host, const std::string& service ){
        // Resolve the host name into an IP address.
        boost::asio::ip::tcp::resolver resolver(io_service);
        boost::asio::ip::tcp::resolver::query query(host, service);
        boost::asio::ip::tcp::resolver::iterator endpoint_iterator =
        resolver.resolve(query);
        boost::asio::ip::tcp::endpoint endpoint = *endpoint_iterator;
        
        // Start an asynchronous connect operation.
        connection_.socket().async_connect(endpoint,
                                           boost::bind(&CommitClass::handle_connect, this,
                                                       boost::asio::placeholders::error/*, ++endpoint_iterator*/));
    }
    
    void CommitClass::handle_connect(const boost::system::error_code& e)
    {
        int seed = 3333;
        
        
        Node* node = verifier->getRBSL()->search(Key(0));
        node->print();
        cout<<"-----------------------"<<endl;
        /*First diffs are generated out of the changed file*/
        vector<DiffSkipList> ad = DiffSkipList::applyDiffToSkiplists(verifier, verifier->getTempDatas(), DPDPCONST::minBlockLength, DPDPCONST::maxBlockLength);
        //cout<<"3"<<endl;
        
        /*And they are sent to the server as partial updates*/
        PacketClient p( PacketBase::COMMIT, seed, ad, getID());
        
        
        
        connection_.async_write(p,
                                boost::bind(&CommitClass::handle_write, this,
                                            boost::asio::placeholders::error, ad, seed));
        
    }
    
    void CommitClass::handle_write(const boost::system::error_code& e, vector<DiffSkipList> ad, int seed/*, connection_ptr conn*/)
    {
        connection_.async_read(packet,
                               boost::bind(&CommitClass::handle_read, this,
                                           boost::asio::placeholders::error, ad, seed));
    }
    
    void CommitClass::handle_read(const boost::system::error_code& e, vector<DiffSkipList> diffs, int seed)
    {
        
        srand(seed);
        /*When we get a series of proofs for the block updates we have sent, 					* using the same updates that we generated already,
         * We verify if the updates are done correctly
         */
        if(packet.getType() == PacketServer::COMMITPROOF)
        {
            cout<<"Packet arrived"<<endl;
            if(!verifier->VerifyMultiProofWOchalIndex(packet.getProofVec(),packet.getTags(), packet.getRewriteProof()))
            {
                
                cout<<"!!!=>>First Proof is not verified"<<endl;
                
                
            }else{
                cout<<"First Proof is verified... "<<endl;
                cout<<"seed:"<<seed<<endl;
                for(int i = 0 ; i<diffs.size();i++)
                {
                    diffs[i].print();
                }
                /*	cout<<"tags"<<endl;
                 for(int i = 0 ; i<packet.getTags().size() ;i++)
                 {
                 cout<<i<<"->"<<packet.getTags()[i]<<endl;
                 }*/
                
                RankBasedSkipList* dummyRbsl = DPDPFlexVerifier::generateDummySkipList(verifier,packet.getProofVec(),packet.getTags());
                SetSeed( seed );
                srand( to_int( seed ) );
                
                DiffSkipList::applyDiffsToSLforVerifier(dummyRbsl, diffs, "clientFolder/aFileOf16000KiB.txt");
                
                cout<<"--*****----"<<endl;
                ZZ rh = dummyRbsl->getRoot()->getHash();
                cout<<"--*****----"<<endl;
                
                ((RankBasedNode*)dummyRbsl->getRoot())->calculateHashRec();
                
                cout<<"--*****----"<<endl;
                dummyRbsl->printRB();
                cout<<"--*****----"<<endl;
                
                
                
                cout<<"The root for rbsl1: "<< packet.getUpdatedHash() <<endl;
                cout<<"The root for dummy: "<< rh <<endl;
                cout<<"The root for dummy2: "<< dummyRbsl->getRoot()->getHash() <<endl;
                
            }
            
        }else{
            cout<<"Wrong packet arrived! (waited: REWRITEPROOF)"<<endl;
        }
        
    }
    
    //END CommitClass:: function definitions
    
    //START TransmissionClass:: function definitions
    
    TransmissionClass::TransmissionClass(boost::asio::io_service& io_service,
                                         const std::string& host, const std::string& service, char* fN, DPDPFlexVerifier* verif, int i)
    : connection_(io_service)
    {
        verifier = verif;
        fileName = fN;
        id = i;
        connect(io_service, host, service);
    }
    
    void TransmissionClass::connect(boost::asio::io_service& io_service, const std::string& host, const std::string& service ){
        // Resolve the host name into an IP address.
        boost::asio::ip::tcp::resolver resolver(io_service);
        boost::asio::ip::tcp::resolver::query query(host, service);
        boost::asio::ip::tcp::resolver::iterator endpoint_iterator =
        resolver.resolve(query);
        boost::asio::ip::tcp::endpoint endpoint = *endpoint_iterator;
        
        // Start an asynchronous connect operation.
        connection_.socket().async_connect(endpoint,
                                           boost::bind(&TransmissionClass::handle_connect, this,
                                                       boost::asio::placeholders::error/*, ++endpoint_iterator*/));
    }
    
    void TransmissionClass::handle_connect(const boost::system::error_code& e)
    {
        vector<ZZ> tags;
        
        verifier->divideAndBuildSkipList_sendFileToServerPL(to_int(verifier->getSeed()), &tags, "clientFolder/aFileOf16000KiB", DPDPCONST::defaultBlockLength);
        
        Data d("clientFolder/aFileOf16000KiB.txt");
        vector<char> b;
        b.push_back('0');
        PacketClient p( PacketClient::TRANSMISSION, b/*d.getBytesFromFile()*/, tags, getID());
        
        p.print();
        connection_.async_write(p,
                                boost::bind(&TransmissionClass::handle_write, this,
                                            boost::asio::placeholders::error));
        cout<<"FILE TRANSMISSION PACKET SENT"<<endl;
        
    }
    
    void TransmissionClass::handle_write(const boost::system::error_code& e/*, connection_ptr conn*/)
    {
        connection_.async_read(packet,
                               boost::bind(&TransmissionClass::handle_read, this,
                                           boost::asio::placeholders::error));
    }
    
    void TransmissionClass::handle_read(const boost::system::error_code& e)
    {
        if(packet.getType() == PacketServer::REWRITEPROOF)
        {
            cout<<"META DATA RECEIVED"<<endl;
            //verifier->verifyRewrite( packet.getRewriteProof() );
            verifier->setMetaData(packet.getRewriteProof())	;
        }else{
            cout<<"Wrong packet arrived! (waited: REWRITEPROOF)"<<endl;
        }
    }
    
    //END TransmissionClass:: function definitions
    
    //START ChallengeClass:: function definitions
    
    ChallengeClass::ChallengeClass(boost::asio::io_service& io_service,
                                   const std::string& host, const std::string& service, char* fileName, DPDPFlexVerifier* verif, int i, int se, int cCount)
    : connection_(io_service)
    {
        id = i;
        verifier = verif;
        sec = se;
        clientCount = cCount;
        connect(io_service, host, service);
    }
    
    void ChallengeClass::connect(boost::asio::io_service& io_service, const std::string& host, const std::string& service ){
        startTimer();
        
        // Resolve the host name into an IP address.
        boost::asio::ip::tcp::resolver resolver(io_service);
        boost::asio::ip::tcp::resolver::query query(host, service);
        boost::asio::ip::tcp::resolver::iterator endpoint_iterator =
        resolver.resolve(query);
        boost::asio::ip::tcp::endpoint endpoint = *endpoint_iterator;
        
        // Start an asynchronous connect operation.
        connection_.socket().async_connect(endpoint,
                                           boost::bind(&ChallengeClass::handle_connect, this,
                                                       boost::asio::placeholders::error/*, ++endpoint_iterator*/));
    }
    
    void ChallengeClass::handle_connect(const boost::system::error_code& e)
    {
        float cTime = printTimer(0, "A challenge of 460 random blocks (Establishing connection)");
        startTimer();
        int chalCount = DPDPCONST::numOfChallenges;
        int defaultBlockLength = DPDPCONST::defaultBlockLength;
        srand ( time(NULL) );
        int seed = rand() % DPDPCONST::seedRange;
        //int seed = to_ZZ(rand() % DPDPCONST::seedRange);
        //int seed = DPDPCONST::Seed; //will be randomly generated
        int chal;
        srand( seed );
        challenges.clear();
        while(challenges.size() < chalCount) {  //for flexlist tries
            chal = (rand() % ((verifier->getSize() - 1) * defaultBlockLength));
            chal = chal - chal % defaultBlockLength;
            vector<int>::iterator it = std::find(challenges.begin(), challenges.end(), chal);
            if (it != challenges.end() || chal == 0)
                continue;
            challenges.push_back(chal);
        }
        
        sort(challenges.begin(), challenges.begin()+chalCount);
        
        PacketClient p( PacketClient::CHALLENGE, chalCount, seed, getID());
        
        Data d("PLTESTDATA/20KBUPDATE.txt");		//SIMULATION
        p.setBlock(d.getBytesFromFile());		//SIMULATION
        
        p.print();
        //				cout<<"CHALLENGE PACKET SENT"<<endl;
        float gTime = printTimer(1, "A challenge of 460 random blocks(chal generation)");
        connection_.async_write(p,
                                boost::bind(&ChallengeClass::handle_write, this,
                                            boost::asio::placeholders::error, challenges, seed));
        startTimer();
        
    }
    /// Handle completion of a write operation.
    void ChallengeClass::handle_write(const boost::system::error_code& e, vector<int> challenges, int seed/*, connection_ptr conn*/)
    {
        connection_.async_read(packet,
                               boost::bind(&ChallengeClass::handle_read, this,
                                           boost::asio::placeholders::error, challenges, seed));
    }
    void ChallengeClass::handle_read(const boost::system::error_code& e, vector<int> challenges, int seed)
    {
        if(packet.getType() == PacketServer::PROOF)
        {
            //
            float tTime = printTimer(2, "A challenge of 460 random blocks(transmission, proof generation, transmission)");
            startTimer();
            //			packet.print();
            long long timeElapsedTags;
            SetSeed(seed);
            bool tagsAndBlockSum = false;
            if(verifier->verifyTags(packet.getTags(), packet.getBlockSum(), &timeElapsedTags ))
                tagsAndBlockSum = true;
            //cout<<"Verifier: TAGS AND BLOCKSUM PROOF VERIFIED"<<endl;
            //else
            //	cout<<"Verifier: TAGS AND BLOCKSUM PROOF NOT VERIFIED"<<endl;
            //verifier->verifyTags
            
            float fTime;
            if(/*tagsAndBlockSum && */verifier->VerifyMultiProof(challenges, packet.getProofVec(), packet.getTags(), packet.getBlockSum()))
            {
                fTime = printTimer(3, "A challenge of 460 random blocks(just verification)");
                cout<<"Verifier: SKIP LIST PROOF VERIFIED"<<endl;
            }else 	cout<<"Verifier: SKIP LIST PROOF NOT VERIFIED"<<endl;
            
            //verifier->VerifyMultiProof
            char log[200];
            sprintf(log,"%f %f %f\n",tTime, fTime, tTime + fTime);
            
            cout<<log<<endl;
            
            char logFileName[50];
            sprintf(logFileName,"%s%d-%d.txt","clientFolder/log", sec, clientCount);
            ofstream logFile;
            logFile.open(logFileName,std::fstream::app);
            logFile << log;
            logFile.close();
            
        }else{
            cout<<"Wrong packet arrived! "<<packet.getType()<<" (waited: PROOF)"<<endl;
        }
    }
    
    //END ChallengeClass:: function definitions
    
    //START ClientDPDPmain:: function definitions
    
    ClientDPDPmain::ClientDPDPmain(boost::asio::io_service& io_service,
                                   const std::string& host, const std::string& service, int i)
    : connection_(io_service)
    {
        id = i;
        theHost = host;
        theService = service;
        connect(io_service, host, service);
    }
    
    void ClientDPDPmain::connect(boost::asio::io_service& io_service, const std::string& host, const std::string& service ){
        // Resolve the host name into an IP address.
        boost::asio::ip::tcp::resolver resolver(io_service);
        boost::asio::ip::tcp::resolver::query query(host, service);
        boost::asio::ip::tcp::resolver::iterator endpoint_iterator =
        resolver.resolve(query);
        boost::asio::ip::tcp::endpoint endpoint = *endpoint_iterator;
        
        // Start an asynchronous connect operation.
        connection_.socket().async_connect(endpoint,
                                           boost::bind(&ClientDPDPmain::handle_connect, this,
                                                       boost::asio::placeholders::error/*, ++endpoint_iterator*/));
    }
    
    /// Handle completion of a connect operation.
    void ClientDPDPmain::handle_connect(const boost::system::error_code& e)
    {
        
        if (!e)
        {
            
            fileName = "clientFolder/aFileOf16000KiB.txt";
            ZZ n,g;//int size;
            fName = "aFileOf16000KiB";
            
            char fname[80] = "";
            strcat(fname, fName.c_str());
            
            Data d(fileName);
            
            
            vector<ZZ> fileBlocks;
            fileBlocks.push_back(d.getValuePers()); // NOW WE ONLY HAVE 1 BLOCK FOR TEST ISSUES
            
            DPDPFlexVerifier verifer(&fileBlocks, &n, &g);
            this->verifier = verifer;
            this->verifier.setSize(DPDPCONST::numOfBlocks); //XXX this must be updated according to the blocks in the system, left this way for test purposes
            
            PacketClient p( PacketClient::INIT, this->verifier.getSize(), fname, n, g, getID());
            p.print();cout<<"INITIALIZATION PACKET SENT"<<endl;
            //Packet p( Packet::INIT, this->verifier.getSize(), fileName, n, g);  //IF THERE IS A METADATA SET BEFOREHAND CLIENT SHOULD SEND REINIT SO THAT SERVER WILL KNOW THIS
            //CLIENT WAS ALREADY SERVED ONCE.
            
            
            // Successfully accepted a new connection. Send the packet to the
            // server. The connection::async_write() function will automatically
            // serialize the data structure for us.
            connection_.async_write(p,
                                    boost::bind(&ClientDPDPmain::handle_write, this,
                                                boost::asio::placeholders::error/*, connection_*/));
            
            // Successfully established connection. Start operation to read the info
            //.The connection::async_read() function will automatically
            // decode the data that is read from the underlying socket.
            connection_.async_read(packet,
                                   boost::bind(&ClientDPDPmain::handle_read, this,
                                               boost::asio::placeholders::error));
            
        }
        else
        {
            // An error occurred. Log it and return. Since we are not starting a new
            // operation the io_service will run out of work to do and the client will
            // exit.
            std::cerr << e.message() << std::endl;
        }
        
    }
    /// Handle completion of a read operation.
    void ClientDPDPmain::handle_read(const boost::system::error_code& e)
    {
        if (!e)
        {
            if(packet.getType() == PacketServer::INFO)
            {
                cout<<"CONNECTION ESTABLISHED"<<endl;
                cout<<"Verifier: INFO packet arrived."<<endl;
                packet.print();
                verifier.setSeed(to_ZZ(packet.getSeed()));
                //cout<<"Seed which is set by INFO packet: "<<verifier.getSeed()<<endl;
                
            }
            
            startOperations(printMenu());
        }
        else
        {
            // An error occurred.
            std::cerr << e.message() << std::endl;
        }
        
        // Since we are not starting a new operation the io_service will run out of
        // work to do and the client will exit.
        cout<<"ee"<<endl;
    }
    
    /// Handle completion of a write operation.
    void ClientDPDPmain::handle_write(const boost::system::error_code& e/*, connection_ptr conn*/)
    {
        // Nothing to do. The socket will be closed automatically when the last
        // reference to the connection object goes away.
    }
    
    
    int ClientDPDPmain::printMenu(){
        int userChoice;
        cout<<endl<<endl<<"What will you do?"<<endl;
        cout<<"1 "<<"Send file(ready)"<<endl;
        cout<<"2 "<<"Challenge file(ready)"<<endl;
        /*cout<<"3 "<<"Modify block(implemenation done, not ready for test)"<<endl;
         cout<<"4 "<<"Insert block(implemenation done, not ready for test)"<<endl;
         cout<<"5 "<<"Delete block(implemenation done, not ready for test)"<<endl;*/
        //cout<<"3 "<<"Commit!"<<endl;
        cout<<"3 "<<"Challenge in a row"<<endl;
        cout<<"4 "<<"Download file(not applicable to current scenario)"<<endl;
        
        cin>> userChoice;
        
        return userChoice;
    }
    
    
    
    
    void ClientDPDPmain::startOperations(int choice){
        
        
        switch (choice)
        {
                
            case 1: {
                try
                {
                    //cout<<"What is the name of the file you want to store?"<<endl;
                    //burada da okunabilir:?
                    cout<<"TRANSMISSION STARTED"<<endl;
                    boost::asio::io_service io_service;
                    s11n_example::TransmissionClass transmission(io_service, theHost, "5195", fileName, &verifier, getID());
                    io_service.run();
                }
                catch (std::exception& e)
                {
                    std::cerr << e.what() << std::endl;
                }
                
                
                break;
            }
            case 2:
                try
            {
                //burada da okunabilir:?
                cout<<"CHALLENGE OF 460 RANDOM BLOCKS STARTED"<<endl;
                boost::asio::io_service io_service;
                s11n_example::ChallengeClass challenge(io_service, theHost, "5195", fileName, &verifier, getID());
                io_service.run();
            }
                catch (std::exception& e)
            {
                std::cerr << e.what() << std::endl;
            }
                
                break;
                /*case 3:
                 
                 try
                 {
                 //burada da okunabilir:?
                 boost::asio::io_service io_service;
                 s11n_example::ModifyClass modify(io_service, "127.0.0.1", "5195", fileName, &verifier);
                 io_service.run();
                 }
                 catch (std::exception& e)
                 {
                 std::cerr << e.what() << std::endl;
                 }
                 
                 break;
                 
                 case 4:
                 
                 try
                 {
                 fileName = "clientFolder/aFileOf16000KiB";
                 boost::asio::io_service io_service;
                 s11n_example::InsertClass insert(io_service, "127.0.0.1", "5195", fileName, &verifier);
                 io_service.run();
                 }
                 catch (std::exception& e)
                 {
                 std::cerr << e.what() << std::endl;
                 }
                 
                 break;
                 */
                /*case 3:
                 
                 try
                 {
                 fileName = "clientFolder/aFileOf16000KiB";
                 boost::asio::io_service io_service;
                 s11n_example::CommitClass commit(io_service, theHost, "5195", fileName, &verifier, getID());
                 io_service.run();
                 }
                 catch (std::exception& e)
                 {
                 std::cerr << e.what() << std::endl;
                 }
                 
                 break;
                 */
            case 3:
                try
            {
                int waitBeforeChal = 10; //sec //not used for fake upload version
                
                int cCount = 20;
                int tInt = 1000000;
                int iterations = 20;
                cout<<"How many client there are?"<<endl;
                cin>>cCount;
                cout<<"What is the time interval?"<<endl;
                cin>>tInt;
                cout<<"How many iterations you wish?"<<endl;
                cin>>iterations;
                usleep(waitBeforeChal * 1000000);
                for(int i = 0; i<iterations; i++){
                    cout<<"CHALLENGE OF 460 RANDOM BLOCKS STARTED"<<endl;
                    boost::asio::io_service io_service;
                    s11n_example::ChallengeClass challenge(io_service, theHost, "5195", fileName, &verifier, getID(), tInt, cCount);
                    io_service.run();
                    usleep(tInt);
                }
            }
                catch (std::exception& e)
            {
                std::cerr << e.what() << std::endl;
            }
                
                break;
            case 99:
                remove( "clientFolder/counter.txt" );
                break;
        }
        
        startOperations(printMenu());
    }
    
    //END ClientDPDPmain:: function definitions
    
} // namespace s11n_example
#endif /*_ClientDPDPmain_H_*/
