#ifndef _CommitProof_H_
#define _CommitProof_H_

#include <vector>
#include <NTL/ZZ.h>
#include <cstdlib>
#include <DPDPCONST.h>
#include <string.h>
#include <Serialize.h>
#include <boost/serialization/vector.hpp>
#include <DiffSkipList.h>
#include "ProofNode.h"

/*------------------------------------------------------------------------
 * Program:   	CommitProof
 *
 * Purpose:   	This class is generated for ease of understanding the proofs generated for a series of updates
 *
 *
 * Note:      	This class is meant to be sent over network for proofs
 *
 *------------------------------------------------------------------------
 */

class CommitProof {
    
public:
    
    CommitProof(vector<ProofNode> p1, vector<ProofNode> pupv1, vector<ZZ> tsB1, vector<ZZ> ts1);
    CommitProof();
    ~CommitProof();
    void print();
    
    friend class boost::serialization::access;
    template <class Archive>
    void serialize(Archive& ar, const unsigned int ver);
    
    vector<ProofNode> getProofVec();
    vector<ProofNode> getProofVecUpdated();
    vector<ZZ> getTagsBefore();
    vector<ZZ> getTagsUpdated();
    
private:
    
    //Proof vector generated before the update
    vector<ProofNode> p;
    //Proof vector generated after the update
    vector<ProofNode> pupv;
    //Tags for the challenged nodes before the update
    vector<ZZ> tsB;
    //Tags for the challenged nodes after the update
    vector<ZZ> ts;
    
};

#endif /*_CommitProof_H_*/

#ifndef _CommitProof_cpp_
#define _CommitProof_cpp_

CommitProof::CommitProof(vector<ProofNode> p1, vector<ProofNode> pupv1, vector<ZZ> tsB1, vector<ZZ> ts1)
: p(p1), pupv(pupv1), tsB(tsB1), ts(ts1){
}

CommitProof::CommitProof() {}

CommitProof::~CommitProof() {}

void CommitProof::print() {}


template <class Archive>
void CommitProof::serialize(Archive& ar, const unsigned int ver) {
    ar  	& auto_nvp(p)
				& auto_nvp(pupv)
				& auto_nvp(tsB)
				& auto_nvp(ts)
    ;
}

vector<ProofNode> CommitProof::getProofVec() { return p; }
vector<ProofNode> CommitProof::getProofVecUpdated() { return pupv; }
vector<ZZ> CommitProof::getTagsBefore() { return tsB; }
vector<ZZ> CommitProof::getTagsUpdated() { return ts; }

#endif /*_CommitProof_cpp_*/