#ifndef _DPDPProver_H_
#define _DPDPProver_H_

#include "DPDPCONST.h"
#include "RankBasedSkipList.h"

class DPDPProver {
public:
	
	DPDPProver(ZZ* seed, ZZ N, ZZ g) {
		srand ( time(NULL) );
		*seed = to_ZZ(DPDPCONST::Seed);
		//*seed = to_ZZ(rand() % DPDPCONST::seedRange);
		setN(N);setG(g);
		
		sl = new RankBasedSkipList(N,g);
	}
	
	~DPDPProver()
	{
		delete sl;
	}
	ZZ fullRewrite(ZZ seed, vector<ZZ> *fileBlocks, vector<ZZ> tags) {
		srand ( to_int(seed) );SetSeed( seed );
		for(int i=1; i <= DPDPCONST::numOfBlocks/*fileBlocks->size()*/; i++) {
			sl->insert(Key(i), Data((*fileBlocks)[0]), tags[0]);//i i
		}
		//sl->print();
		return sl->getRoot()->getHash();
	}
	
	ZZ genBlockSum (ZZ seed, vector<ZZ> fileBlocks, int stat) {
		ZZ randValue;
		SetSeed(seed);
		ZZ blockSum = 0;	
		
		for(int i = 0 ; i < fileBlocks.size(); i++)
		{
			RandomLen(randValue, stat);
			//cout<<"randVal in genBlockSum: "<< randValue<<endl;	
			blockSum += randValue * fileBlocks[i];
		}
		return blockSum;
	}	
	
	Node* SearchMulti(Node* currentNode, vector<int> chalIndex, int start, int end, int rankState, ZZ N, ZZ g, vector<ProofNode>* proofVector, stack<State*>* interStack) {
		Key key(chalIndex[start]-rankState);
		bool breakPoint = true;
		bool canGoDown;
		int rankDiff;
		int tempEnd;
		while(breakPoint){
			ProofNode proofNode(currentNode->getLevel(), currentNode->getKey().getValue());
			if (key.getValue() == 0 && currentNode->getLevel() == 0) { 
				breakPoint = false;
				proofNode.setEndFlag();
			}		
			canGoDown = currentNode->canGoDown(key, -1);

			rankDiff = currentNode->getBelowKeyValue();
			tempEnd = end;
			if (currentNode->isIntersecting(canGoDown, chalIndex, key.getValue(), &end, rankState)) {
				
				interStack->push(new State(*currentNode, tempEnd, rankState));
				proofNode.setInterFlag();			
			}
			
			if (canGoDown) {
				proofNode.set(currentNode->getAfterHash(), false);
				if(breakPoint)
					currentNode = currentNode->getBelow();		
			}
			else if(breakPoint){
				proofNode.set(currentNode->getBelowHash(N,g), true);
				key.setValue(key.getValue() - currentNode->getBelowKeyValue());
				rankState = rankState + currentNode->getBelowKeyValue();
				currentNode = currentNode->getAfter();
			}
			proofVector->push_back(proofNode);

		}
		return currentNode;
	}
	
	vector<ProofNode> GenMultiProof(vector<int> chalIndex, vector<ZZ>* tags, vector<ZZ>* fileBlocks) {
		Node* currentNode = sl->getRoot();
		stack<State*> interStack;
		vector<ProofNode> proofVector;
		State* state = new State(*currentNode, chalIndex.size()-1, 0); 
		int rankState = 0;
		for(int i=0; i < chalIndex.size(); i++) {
			currentNode = SearchMulti(currentNode, chalIndex, i, state->getEndIndex(), state->getRankState(), N, g, &proofVector, &interStack);
			tags->push_back(currentNode->getTag(N,g));
			fileBlocks->push_back(currentNode->getData().getValue());
			if(!interStack.empty()){
				state = interStack.top(); interStack.pop();
				currentNode = state->getNode().getAfter();
				state->setRankState(state->getRankState() + state->getNode().getBelowKeyValue());
			}			
		}
		delete state;
		return proofVector;
	}
	
	
	void update(int index, bool type, int level = 0, ZZ tag = to_ZZ(0), ZZ block = to_ZZ(0)){ //if type is true then insert, otherwise delete
		if (type)
			sl->insert(Key(index), Data(block), tag, NULL, level); 
		else sl->remove(Key(index), NULL);
	}
		
	
	ZZ getN(){return N;}
	void setN(ZZ n){ N = n;}
	ZZ getG(){return g;}
	void setG(ZZ G){ g = G;}
	RankBasedSkipList* getSL(){return sl;}
	
private:
	ZZ N,g;
	RankBasedSkipList *sl;
};

#endif /* _DPDPProver_H_ */
