#ifndef _DPDPVerifier_H_
#define _DPDPVerifier_H_

#include "GroupRSA.h"
#include "Timer.h"
#include "RankBasedSkipList.h"
#include "MultiExp.h"

class DPDPVerifier {
public:
	DPDPVerifier(vector<ZZ> *Data, ZZ *N, ZZ *g)
	{
		data = Data;
		srand ( time(NULL) );
		generateGroup(N, g);
	}
	
	~DPDPVerifier()
	{
		data->clear();
		//delete data;
	}
	
	void generateGroup(ZZ *N, ZZ *g) {
		ZZ returnN, returng;
		SetSeed( to_ZZ(rand() % 10000) );	
		
		GroupRSA* group = new GroupRSA("Client1", 1024, 80);
		returnN = group->getModulus();
		returng = RandomBnd(returnN);
		while (GCD(returng, returnN) != 1) {
			returng = RandomBnd(returnN);
		}
		setN(returnN);
		setG(returng);
		setP(group->getP());
		setQ(group->getQ());
		*N = returnN; *g= returng;
	}
	
	long long sendFileToServer(ZZ seed, vector<ZZ> *tags)
	{		
		//SetSeed( seed );
		RankBasedSkipList *sl = new RankBasedSkipList(N, g);
		
		srand ( to_int(seed) );
		ZZ tag;
		ZZ MD;
		long long timer = 0;
		cout<<DPDPCONST::numOfBlocks<<endl;
////////////////
			startTimer();
			tag = calculateTag((*data)[0], N, g, p, q);//j
			timer += elapsedTime();
			tags->push_back(tag);
///////////////////////
		for(int j=1; j<= DPDPCONST::numOfBlocks; j++) {
//////////////////////
///bu araya gelecek
/////////////////////

			sl->insert(Key(j), Data((*data)[0]), tag);//j
		}
		MD = sl->getRoot()->getHash();
		//sl->print();
		delete sl;
		
		MetaData = MD;
		return timer/1000.0;
	}
	
	bool verifyRewrite(ZZ rootRewrite) {
		if(MetaData == rootRewrite) {
			data->clear(); 
			cout<<"Verifier : VERIFIED, DELETED DATA"<<endl;
			return true;
		}

		cout<<"Verifier : NOT VERIFIED"<<endl;
		return false;
	}
	
	ZZ calculateTag(ZZ data, ZZ modulus, ZZ element, ZZ p, ZZ q) { 
		ZZ exp = to_ZZ(1); 
		mpz_mod(MPZ(exp), MPZ(data), MPZ((to_ZZ((p-1) * (q-1)))));
		//exp = (data % (to_ZZ((p-1) * (q-1))));
		return PowerMod(element, exp, modulus);	 
	}
	
	vector<int> generateChallenges(int nChallenges, int numBlocks, ZZ* seed) {	
		*seed = to_ZZ(rand() % 10000);
		this->seed = *seed;
		vector<int> toReturn;
		while(toReturn.size() < nChallenges) {
			int chalx = NTL::RandomBnd(numBlocks);//rand() % numBlocks;
			vector<int>::iterator it = std::find(toReturn.begin(), toReturn.end(), chalx);
			if (it != toReturn.end() || chalx == 0)
				continue;
			toReturn.push_back(chalx);
		}
		sort(toReturn.begin(), toReturn.begin()+nChallenges);
		
		//toReturn.push_back(2);toReturn.push_back(3);toReturn.push_back(7);toReturn.push_back(8);toReturn.push_back(9);
		return toReturn;
	}
	
	
	bool verifyTags(vector<ZZ> tags, ZZ blockSum, long long* timer) {
		ZZ randValue;
		ZZ tagMul;
		ZZ result;
		tagMul = 1;
		SetSeed(seed);
		startTimer();
		
		vector<ZZ> rndVals;
		for(int i = 0 ; i < tags.size(); i++)
		{
			RandomLen(randValue, DPDPCONST::stat);
			rndVals.push_back(randValue);
			//cout<<"randVal in verlifyTags: "<< randValue<<endl;
		}
		
		tagMul = MultiExp(tags, rndVals, 
			getN(), false);
		//for(int i = 0 ; i < tags.size(); i++)
		//{
			//RandomLen(randValue, DPDPCONST::stat);
			//tagMul *= PowerMod(tags[i], randValue, getN());	
		//}
		
		
		mpz_mod(MPZ(tagMul), MPZ(tagMul), MPZ(getN()));
		//tagMul = to_ZZ(tagMul % getN());
		*timer = elapsedTime()/1000.0;
		mpz_mod(MPZ(blockSum), MPZ(blockSum), MPZ((to_ZZ((p-1) * (q-1)))));
		//blockSum = (blockSum % (to_ZZ((p-1) * (q-1))));
		result = PowerMod(getG(), blockSum, getN());

		//cout<<tagMul<<endl<<endl;
		//cout<<result<<endl<<endl;

		if(tagMul == result) return true;
		return false;
	}
	
	void updateRankSum(int rankDiff, vector<int>* vecRankSum, int start, int end) {
		for(int i=start; i<=end; i++)
			(*vecRankSum)[i] += rankDiff;
	}
	
	bool VerifyMultiProof(vector<int> chalIndex, vector<ProofNode> proofVector, vector<ZZ> tags, ZZ blockSum) {
		int start= tags.size() - 1; int end =start;
		int tagIndex = tags.size()-1;
		vector<int> vecRankSum(chalIndex.size(), 0);
		ZZ hash, hashOfPrev;
		int rankOfPrev = 0;
		stack<ZZ> interHashStack;
		stack< pair<int,int> > index; 
		for(int i=proofVector.size()-1; i > -1 ; i--) {
			if(proofVector[i].isEnd() && proofVector[i].isInter()) {
				hash = ProofNode::calculateHash(proofVector[i].getLevel(), proofVector[i].getRank(), tags[tagIndex--], hashOfPrev);
				updateRankSum(1, &vecRankSum, start, end);
				start--;
			}
			else if(proofVector[i].isEnd()) {
				if(tagIndex != tags.size()-1) { 
					interHashStack.push(hashOfPrev);
					index.push(pair<int,int>(start, end));
					start--; 
					end = start;
				}
				hash = ProofNode::calculateHash(proofVector[i].getLevel(), proofVector[i].getRank(), tags[tagIndex--], proofVector[i].getHash());
				
			}
			else if(proofVector[i].isInter()) {
				end = index.top().second;
				updateRankSum(rankOfPrev, &vecRankSum, index.top().first, index.top().second); index.pop();
				hash = ProofNode::calculateHash(proofVector[i].getLevel(), proofVector[i].getRank(), hashOfPrev,  interHashStack.top()); 
				interHashStack.pop();
			}
			else if(proofVector[i].getRgtOrDwn() || proofVector[i].getLevel() == 0) {
				hash = ProofNode::calculateHash(proofVector[i].getLevel(), proofVector[i].getRank(), proofVector[i].getHash(), hashOfPrev);
				updateRankSum(proofVector[i].getRank() - rankOfPrev, &vecRankSum, start, end);
			}
			else {
				hash = ProofNode::calculateHash(proofVector[i].getLevel(), proofVector[i].getRank(), hashOfPrev, proofVector[i].getHash());
			}
			hashOfPrev = hash;
			//cout << "multihash:"<<hash<<endl;
			rankOfPrev = proofVector[i].getRank();
		}
		return ((hash == MetaData) && equal(chalIndex.begin(),chalIndex.end() , vecRankSum.begin()));
	}
	
	bool VerifyInsert(vector<int> chalIndex, vector<ProofNode> proofVector, vector<ZZ> tags, ZZ blockSum, ZZ tag, int level) {
		if(!VerifyMultiProof(chalIndex, proofVector, tags, blockSum)){
			cout << "wtf"<<endl;
			return false;
		}
		else {
			int rank = 1, i = proofVector.size() - 1, rankTower = proofVector[i].getRank();
			ZZ hash = tags[0], hashTower = ProofNode::calculateHash(0, rankTower, tag, proofVector[i].getHash());
			if (level != 0){
				hash = ProofNode::calculateHash(0, 1, hash, to_ZZ(0)); // is last argument right?
			}
			i--;
			while (proofVector[i].getLevel() < level || (proofVector[i].getLevel() == level && !proofVector[i].getRgtOrDwn()) ) {
				if (proofVector[i].getRgtOrDwn()) {
					rank += proofVector[i].getRank() - proofVector[i+1].getRank();
					hash = ProofNode::calculateHash(proofVector[i].getLevel(), rank, proofVector[i].getHash(), hash);
				}
				else {
					rankTower += proofVector[i].getRank() - proofVector[i+1].getRank();
					hashTower = ProofNode::calculateHash(proofVector[i].getLevel(), rankTower, hashTower, proofVector[i].getHash());
				}
				i--;
			}
			hash = ProofNode::calculateHash(level, rank + rankTower, hash, hashTower);
			MetaData = calculateRemainingHashes(i, hash, 1, proofVector);
			return true;
		}
	}
	
	/*bool VerifyInsert2(vector<int> chalIndex, vector<ProofNode> proofVector, vector<ZZ> tags, ZZ blockSum, ZZ tag, int level) {
		if(!VerifyMultiProof(chalIndex, proofVector, tags, blockSum)){
			cout << "wtf"<<endl;
			return false;
		}
		else {
			stack<ProofNode*>* = s;
			int i = 0;
			s->push(&(proofVector[0]));
			nextPosP(proofVector, &i, true, level, s);
			if (proofVector[i].getRgtOrDwn() || proofVector[i].getLevel() != level){
				newNode = ProofNode(level);
				s->push(&newNode);
			}
			i++;
			while (!proofVector[i].isEnd()){
					*/
	
		
	 ZZ calculateRemainingHashes(int i, ZZ hash, int rankDiff, vector<ProofNode> proofVector) { // is proofVector necessary?
		 //cout << "calculateRemainingHashesStart"<<endl;
		while (i >= 0) {
			if (proofVector[i].getRgtOrDwn()) {
				hash = ProofNode::calculateHash(proofVector[i].getLevel(), proofVector[i].getRank() + rankDiff, proofVector[i].getHash(), hash);
			}
			else {
				hash = ProofNode::calculateHash(proofVector[i].getLevel(), proofVector[i].getRank() + rankDiff, hash, proofVector[i].getHash());
			}
			//cout << hash<<endl;
			//cout << "i:"<<i<< " hash: "<<hash<<endl;
			i--;
		}
		 //cout << "calculateRemainingHashesEnd";
		return hash;
	}
	
	bool VerifyDelete(vector<int> chalIndex, vector<ProofNode> proofVector, vector<ZZ> tags, ZZ blockSum) {
		if(!VerifyMultiProof(chalIndex, proofVector, tags, blockSum)){
			cout << "wtf"<<endl;
			return false;
		}
		else {
			//cout << "VerifyDelete"<<endl;
			int dn = proofVector.size() - 1, i = proofVector.size() - 2;
			while (!proofVector[i].isEnd()) {
				i--;
				//cout << "i:"<<i<<endl;
			}
			int rank = proofVector[dn].getRank();
			ZZ hash = ProofNode::calculateHash(0, rank, tags[0], proofVector[dn].getHash());
			//cout << "i:"<<i<< " hash: "<<hash<<endl;
			dn--;
			//cout << "dn:"<<dn<<endl;
			//cout << dn<<endl;
			if (!proofVector[dn].isEnd() || !proofVector[i].isInter()){
				//cout << "i:"<<i<< " hash: "<<hash<<endl;
				i--;
			}
			//cout << "enterWhile"<<endl;
			while (!proofVector[dn].isEnd() || !proofVector[i].isInter()){
				if (proofVector[i].getLevel() < proofVector[dn].getLevel() || proofVector[dn].isEnd()) {
					rank += proofVector[i].getRank() - proofVector[i+1].getRank();
					hash = ProofNode::calculateHash(proofVector[i].getLevel(), rank, proofVector[i].getHash(), hash);
					//cout << "i:"<<i<< " hash: "<<hash<<endl;
					i--;
				}
				else {
					rank += proofVector[dn].getRank() - proofVector[dn+1].getRank();
					hash = ProofNode::calculateHash(proofVector[dn].getLevel(), rank, hash, proofVector[dn].getHash());
					//cout << "dn:"<<dn<< " hash: "<<hash<<endl;
					dn--;
					
				}
				//cout << "i:"<< i << " dn: " <<dn<<endl;
			}
			//cout << "leaveWhile"<<endl;
			i--;
			//cout << "i:"<<i<<endl;
			MetaData = calculateRemainingHashes(i, hash, -1, proofVector);
			return true;
		}
	}
	
	/*bool DPDPVerifier::nextPosP(vector<ProofNode> proofVector, int* i, bool npi, int levelToInsert,stack<ProofNode*>* s) {
		while ((proofVector[i].getRgtOrDwn() && proofVector[i+1].getLevel >= levelToInsert) || (!proofVector[i].getRgtOrDwn() && proofVector[i+1].getLevel >= levelToInsert && npi)){
			(*i)++;
			s->push(&(proofVector[*i]));
		}
		return true;
	}*/
	

	
	
	ZZ getN(){return N;}
	void setN(ZZ n){ N = n;}
	ZZ getG(){return g;}
	void setG(ZZ G){ g = G;}
	ZZ getP(){return p;}
	void setP(ZZ P){ p = P;}
	ZZ getQ(){return q;}
	void setQ(ZZ Q){ q = Q;}
private:
	ZZ N,g,p,q, MetaData, seed;
	vector<ZZ> *data;
	
};

#endif /* _DPDPVerifier_H_ */
