#ifndef _DiffSkipList_H_
#define _DiffSkipList_H_

#include <NTL/ZZ.h>
#include <DiffBasic.h>
#include <CommitProof.h>
#include <ServerDPDPmain.h>

using namespace std;

/*------------------------------------------------------------------------
 * Program:   	DiffSkipList
 *
 * Purpose:   	This class takes the output of the diff algorithm used and turns it into block updates
 *
 *
 *
 * Note:      	This algorithm is written for test purposes now and it is not finalized yet,
 *				But a prototype, bugs are possible be encountered in a practical usage
 *
 * Functions:
 *				applyDiffToSkiplists: The main method in which we generate the block diff operations for skip list
 *				applyBlockDiffs: This function applies the DiffSkipLists provided to the Flexlist
 *				generateDiffsTest: This function is here for later referance
 *				generateBlocks: With this function we aim to divide the string that we will fit in the skip list
 *								into blocks with at least min block size and at max the maximum block size
 *
 *------------------------------------------------------------------------
 */

class DiffSkipList {
    
public:
    
    enum TYPE { MODIFY = 1, INSERT, REMOVE};
    
    DiffSkipList();
    
    DiffSkipList(TYPE t, int idx, vector<char> d);
    
    DiffSkipList(TYPE t, int idx, DPDPFlexVerifier* verifier, string d = "");
    DiffSkipList(TYPE t, int idx, DPDPFlexVerifier* verifier, vector<char> v, ZZ tag);
    DiffSkipList(TYPE t, int idx, DPDPFlexVerifier* verifier, Data* da, ZZ tag);
    
    DiffSkipList(TYPE t, int idx, string d = "") ;
    /*The main method in which we generate the block diff operations for skip list*/
    static vector<DiffSkipList> applyDiffToSkiplists(DPDPFlexVerifier* verifier, vector<Data>* inVec, int lmin = 10, int lmax = 20);
    /*This function applies the DiffSkipLists provided to the Flexlist*/
    static bool applyBlockDiffs(RankBasedSkipList* rbsl, vector<DiffSkipList> diffs, string fileName, ZZ seed, vector<CommitProof>* proofs = NULL, DPDPFlexProver* prover = NULL);
    static vector<DiffSkipList> generateDiffs(DPDPFlexVerifier* verifier, RankBasedSkipList* rbsl, int csi, int ol, string inString, int lmin = 10, int lmax = 20);
    /*This function is here for later referance*/
    static vector<DiffSkipList> generateDiffsTest(RankBasedSkipList* rbsl, int csi, int ol, string inString, int lmin = 10, int lmax = 20); //unnec
    //generate indices for proof
    static vector<int> indicesToGenerateFirstProofFor(vector<DiffSkipList> diffs);
    static void applyDiffsToSL(DPDPFlexProver* prover,RankBasedSkipList* rbsl,vector<DiffSkipList> diffs,string fileName);
    static void applyDiffsToSLforVerifier(RankBasedSkipList* dummyRbsl,vector<DiffSkipList> diffs,string fileName);
    /*With this function we aim to divide the string that we will fit in the skip list into blocks with at least min block size and at max the maximum block size*/
    static vector<string> generateBlocks(string header, string str, string footer, int lmin, int lmax, bool* tooShort);
    void print();
    
    friend class boost::serialization::access;
    template <class Archive>
    void serialize(Archive& ar, const unsigned int ver);
    
    TYPE getType() {return type;}
    int getIndex();
    vector<char> getData();
    ZZ getTag();
    Data* getDatap();
    
private:
    
    TYPE type;			//Type of the block operation
    int index;			//To which block this will get applied
    vector<char> data;	//The data if this is an insert operation or a modify operation
    Data* d;		//for HashMultiTests
    ZZ tagOfData;		//The tag previously calculated by the client
    
};





//Method and Function definirions//


DiffSkipList::DiffSkipList(){}

DiffSkipList::DiffSkipList(TYPE t, int idx, vector<char> d) {
    type = t;
    index = idx;
    data = d;
    tagOfData = to_ZZ(1);
}

DiffSkipList::DiffSkipList(TYPE t, int idx, DPDPFlexVerifier* verifier, string d) {
    type = t;
    index = idx;
    vector<char> v( d.begin(), d.end() );
    data = v;
    
    ZZ value;
    ZZFromBytes(value, (unsigned char*)d.c_str(), d.size());
    tagOfData = verifier->calculateTag(value, verifier->getN(), verifier->getG(), verifier->getP(), verifier->getQ());
    
    cout << "tag calculated from string: " << tagOfData << endl;
    cout << "------------*****-------------" << endl;
    
    
    //This was to test and see if writing text to a file and then rereading it changes any byte which may couse a change in tag value!!!!!!!!
    //		vector<char> ve( d.begin(), d.end() );
    //		Data dat(ve,"clientFolder/tagCalculationTempFile.txt");
    //		long size = -1;
    //		ZZ tempTag = verifier->calculateTag(dat.getValuePersByStringName(), verifier->getN(), verifier->getG(), verifier->getP(), verifier->getQ());
    //		cout << "tag calculated from File: " << tempTag << endl;
    //		cout << "------------*****-------------" << endl;
}
DiffSkipList::DiffSkipList(TYPE t, int idx, DPDPFlexVerifier* verifier, vector<char> v, ZZ tag) {
    type = t;
    index = idx;
    //vector<char> v( d.begin(), d.end() );
    data = v;
    
    ZZ value;
    //ZZFromBytes(value, (unsigned char*)d.c_str(), d.size());
    tagOfData = tag;//= verifier->calculateTag(value, verifier->getN(), verifier->getG(), verifier->getP(), verifier->getQ());
    
    cout << "tag calculated from string: " << tagOfData << endl;
    cout << "------------*****-------------" << endl;
    
}

DiffSkipList::DiffSkipList(TYPE t, int idx, DPDPFlexVerifier* verifier, Data* da, ZZ tag) {
    type = t;
    index = idx;
    //vector<char> v( d.begin(), d.end() );
    d = da;
    
    ZZ value;
    //ZZFromBytes(value, (unsigned char*)d.c_str(), d.size());
    tagOfData = tag;//= verifier->calculateTag(value, verifier->getN(), verifier->getG(), verifier->getP(), verifier->getQ());
    
    //cout << "tag calculated from string: " << tagOfData << endl;
    //cout << "------------*****-------------" << endl;
    
}

DiffSkipList::DiffSkipList(TYPE t, int idx, string d) {
    type = t;
    index = idx;
    vector<char> v( d.begin(), d.end() );
    data = v;
    tagOfData = to_ZZ(1); // no verifier provided... for old tests
}
/*The main method in which we generate the block diff operations for skip list*/
vector<DiffSkipList> DiffSkipList::applyDiffToSkiplists(DPDPFlexVerifier* verifier, vector<Data>* inVec, int lmin, int lmax)
{
    RankBasedSkipList* outSL = verifier->getRBSL();
    string in = "";
    string out = "";
    Node* ilk = outSL->search(Key(0));
    
    //cout<<"geldim"<<endl;
    
    //outSL->print();
    //cout<<"-----------"<<endl;
    
    while(ilk->getLeafAfter() != NULL)
    {
        //ilk->print();
        
        long size = -1;
        char* cat = ilk->getData().getDataVal(&size);
        string catS  = string(cat);
        //this trim is a must since there is a problem with char pointers in c++ when you return!!!
        string append = catS.substr(0,size);
        out += append;
        ilk = ilk->getLeafAfter();
    }
    for(int k = inVec->size() -1 ; k > -1; k--)
    {
        long size = -1;
        char* cat = (*inVec)[k].getDataVal(&size);
        string catS  = string(cat);
        string append = catS.substr(0,size);
        in += append;
    }
    
    vector<DiffBasic> allDiffs;
    
    allDiffs = DiffBasic::generateDiffs(out, in);
    
    for(int idx = 0; idx < allDiffs.size(); idx++)
    {
        allDiffs[idx].print();
    }
    
    vector<DiffSkipList> allDiffsBlockFormat;
    
    for(int j=allDiffs.size() -1; j>-1; j--)
    {
        
        vector<DiffSkipList> dslVec = DiffSkipList::generateDiffs(verifier, outSL, allDiffs[j].getIndex(), allDiffs[j].getOutLength(), allDiffs[j].getInString(), lmin, lmax);
        
        for(int i = 0 ; i < dslVec.size() ; i++)
        {
            //push all necessary block operations to allDiffs which will be sent to the server
            allDiffsBlockFormat.push_back(dslVec[i]);
            
        }
        
        //SERVER with a not reversed for loop: Apply diffs to the skip list and print the skip list again to see if it is the same with the altered string
        
    }
    
    return allDiffsBlockFormat;
}

/*This function applies the DiffSkipLists provided to the Flexlist*/
bool DiffSkipList::applyBlockDiffs(RankBasedSkipList* rbsl, vector<DiffSkipList> diffs, string fileName, ZZ seed, vector<CommitProof>* proofs, DPDPFlexProver* prover)
{
    SetSeed( seed );srand ( to_int(seed) );
    cout<<"apply block diffs seed: "<<seed<<endl;
    int blockCounter=0;//XXX:will be read from file and be written to file at the end
    //		blockCounter = Data::readCounterInt("DiffTestFiles/counter.txt"); /////////////////////// WHEN TEST!!!
    
    char fileCounter[80]="connectedClients/";
    cout<<"debug0"<<endl;
    if(prover != NULL){
        strcat(fileCounter, prover->getClientIP().c_str());
        strcat(fileCounter, "/counter.txt");
    }
    cout<<"debug1"<<endl;
    cout<<"-----------file name:"<<fileName<<endl;
    if(fileName[0] == 'c' && fileName[1] == 'o'){
        cout<<"server"<<endl;
        blockCounter = Data::readCounterInt(fileCounter/*"serverFolder/counter.txt"*/);
    }else if(fileName[0] == 'c' && fileName[1] == 'l')
    {
        cout<<"client"<<endl;
        blockCounter = Data::readCounterInt("clientFolder/counter.txt");
    }else{
        blockCounter = Data::readCounterInt("DiffTestFiles/counter.txt");
    }
    
    for(int i = 0; i< diffs.size(); i++,blockCounter++)
    {
        char fileNametxt[80] = "";
        strcat(fileNametxt, fileName.c_str());
        Key key(diffs[i].getIndex());
        
        /*if(diffs[i].getType() == DiffSkipList::REMOVE)
         {
         
         rbsl->remove(key);
         }*/
        
        if(diffs[i].getType() == DiffSkipList::MODIFY)
        {
            
            vector<ProofNode> p; vector<ProofNode> pupv; vector<ZZ> tsB; vector<ZZ> ts; vector<ZZ> chalBlocks; vector<int> challenges;
            
            if(proofs != NULL){
                challenges.push_back(diffs[i].getIndex());
                vector<ZZ> chalBlocks;
                p = prover->GenMultiProof(challenges, &tsB, &chalBlocks/*,"serverFolder/aaaBlockNotDividedYet"*//*fileName*/);
                
            }
            
            char str[100];sprintf(str, "%d", blockCounter);strcat(fileNametxt, str);strcat(fileNametxt, "NEW.txt");
            Data d(diffs[i].getData(),fileNametxt,blockCounter);
            rbsl->modify(key, d, diffs[i].getTag());
            
            //cout<<"^+%^+%^+%^+%"<<endl;
            //diffs[i].print();
            //cout<<"^+%^+%^+%^+%"<<endl;
            if(proofs != NULL){
                pupv = prover->GenMultiProof(challenges, &ts, &chalBlocks/*, "serverFolder/aaaBlockNotDividedYet"*//*fnamebaseTemp*/);
                CommitProof cp(p,pupv,tsB,ts);
                proofs->push_back(cp);
                p.clear();pupv.clear();chalBlocks.clear();
            }
        }
        
        if(diffs[i].getType() == DiffSkipList::INSERT)
        {
            
            vector<ProofNode> p; vector<ProofNode> pupv; vector<ZZ> tsB; vector<ZZ> ts; vector<ZZ> chalBlocks; vector<int> challenges;
            
            if(proofs != NULL){
                challenges.push_back(diffs[i].getIndex()-1);
                vector<ZZ> chalBlocks;
                p = prover->GenMultiProof(challenges, &tsB, &chalBlocks/*,"serverFolder/aaaBlockNotDividedYet"*//*fileName*/);
                
            }
            int level=0;
            while((rand() % 2) == 1) {level++;}
            
            char str[100];
            sprintf(str, "%d", blockCounter);
            strcat(fileNametxt, str);
            strcat(fileNametxt, "NEW.txt");
            Data d(diffs[i].getData(),fileNametxt,blockCounter);
            rbsl->insert(key, d, diffs[i].getTag());
            
            if(proofs != NULL){
                challenges.clear();challenges.push_back(diffs[i].getIndex());
                pupv = prover->GenMultiProof(challenges, &ts, &chalBlocks/*, "serverFolder/aaaBlockNotDividedYet"*//*fnamebaseTemp*/);
                CommitProof cp(p,pupv,tsB,ts);
                proofs->push_back(cp);
                p.clear();pupv.clear();chalBlocks.clear();
            }
            
        }
        
        if(diffs[i].getType() == DiffSkipList::REMOVE)
        {
            
            vector<ProofNode> p; vector<ProofNode> pupv; vector<ZZ> tsB; vector<ZZ> ts; vector<ZZ> chalBlocks; vector<int> challenges;
            cout<<"ad[i].getIndex()-----------"<<diffs[i].getIndex()<<endl;
            if(proofs != NULL){
                challenges.clear();
                challenges.push_back(diffs[i].getIndex()-1);
                challenges.push_back(diffs[i].getIndex());
                
                
                p = prover->GenMultiProof(challenges, &tsB, &chalBlocks/*,"serverFolder/aaaBlockNotDividedYet"*//*fileName*/);
            }
            rbsl->remove(key);
            
            if(proofs != NULL){
                challenges.clear();
                //	chalBlocks.clear();
                challenges.push_back(diffs[i].getIndex());//XXX index arrangement have to be made;
                pupv = prover->GenMultiProof(challenges, &ts, &chalBlocks/*, "serverFolder/aaaBlockNotDividedYet"*//*fnamebaseTemp*/);
                cout<<"tags after-----------"<<ts[0]<<endl;
                CommitProof cp(p,pupv,tsB,ts);
                proofs->push_back(cp);
                p.clear();pupv.clear();chalBlocks.clear();
            }
            
        }
    }
    
    //remove the counter
    ofstream counterFile;
    if(fileName[0] == 's')
    {//server
        cout<<"server"<<endl;
        remove(fileCounter /*"serverFolder/counter.txt"*/);
        counterFile.open(fileCounter/*"serverFolder/counter.txt"*/);
    }else if(fileName[0] == 'c')
    {//client
        cout<<"client"<<endl;
        remove( "clientFolder/counter.txt");
        counterFile.open("clientFolder/counter.txt");
    }else
    {//test
        remove( "DiffTestFiles/counter.txt" );
        counterFile.open("DiffTestFiles/counter.txt");
    }
    //counterFile.open("DiffTestFiles/counter.txt");
    counterFile << blockCounter;
    counterFile.close();
}

vector<DiffSkipList> DiffSkipList::generateDiffs(DPDPFlexVerifier* verifier, RankBasedSkipList* rbsl, int csi, int ol, string inString, int lmin, int lmax)
{
    int changeStartIndex = csi;// -1;
    int outLength = ol;// +1;
    vector<DiffSkipList> updateStack;
    vector<DiffSkipList> shortStack;
    int inLength = inString.size();
    Key k(changeStartIndex);
    Node* cN = rbsl->search(&k);
    int remainder = k.getValue() -1;
    /////////////For "short inString" diff operaions to append the new string to the precedent node
    int indexInsidePrecedentNode = changeStartIndex - k.getValue() -1;
    Key key(indexInsidePrecedentNode);
    Node* precedentNode;
    if(indexInsidePrecedentNode >-1)
        precedentNode = rbsl->search(&key);
    //////////////
    cout<<"changeStartIndex: "<<changeStartIndex<<endl;
    
    string header = string(cN->getData().getDataVal()).substr(0,k.getValue()-1); //was k-1
    
    cout << "header: " << header << endl;
    
    
    int outNodeIndex = changeStartIndex - k.getValue();
    //cout << "outNodeIndex: " << outNodeIndex << endl;
    
    vector<int> outStack;
    outStack.push_back(outNodeIndex);
    
    int outCounted = cN->getData().getLength() - remainder;
    //cout << "outCounted: " << outCounted << endl;
    ///////////////////////////////////////////////////////////////////////////////////// List of node indices
    while(outCounted < outLength)
    {
        //cout << "outNodeIndex: " << outNodeIndex << endl;
        outNodeIndex += cN->getData().getLength();
        //cout << "outNodeIndex: " << outNodeIndex << endl;
        //cout << "cN->getData().getLength(): " << cN->getData().getLength() << endl;
        outStack.push_back(outNodeIndex);
        cout<<"PUSH TO OUT STACK ::::::::::::::" <<outNodeIndex<<endl; //XXX: This may catch a bug later!!!
        cN = cN->getLeafAfter();
        //cout << "outCounted: " << outCounted << endl;
        outCounted += cN->getData().getLength();
        //cout << "outCounted: " << outCounted << endl;
        //cout << "cN->getData().getLength(): " << cN->getData().getLength() << endl;
    }
    Node* rightNode = cN->getLeafAfter();
    int insertIndex = outNodeIndex + cN->getData().getLength();
    ///////////////////////////////////////////////////////////////////////////////////// END List of node indices
    long size = -1;
    //cout << "insertIndex: " << insertIndex << endl;
    
    string d = cN->getData().getDataVal(&size);
    //string catS  = string(cat);
    string data = d.substr(0,size);
    
    //cout << "cN.data.size: " << size << endl;
    //cout << "cN.data: " << data << endl;
    
    string footer = data.substr(data.size() - (outCounted - outLength), (outCounted - outLength));
    
    cout << "footer: " << footer << endl;
    
    bool isTooShort = false;
    vector<string> dataStack = DiffSkipList::generateBlocks(header, inString, footer, lmin, lmax, &isTooShort);
    cout<<"isTooShort: " << isTooShort <<endl;
    
    
    if(isTooShort)
    {
        // an index which is inside the precedent node of the node where change started
        // if that index is below 0 that means the change started from the first node, then we need to add the new short block to
        // a node further
        string toDivide;
        int precedentNodeStartIndex = 0;
        if(indexInsidePrecedentNode > 0){
            
            precedentNodeStartIndex = indexInsidePrecedentNode - key.getValue();
            long size = -1;
            string precedentData = precedentNode->getData().getDataVal(&size);
            precedentData = precedentData.substr(0,size);
            //string newDataToBlocks = precedentData + header + inString + footer;
            toDivide = precedentData + dataStack[0];
            
            
            //return updateStack;
        }else{
            //if update occurs at the first block and the inString is short
            //NOT IMPLEMENTED YET!!!
            long size = -1;
            string rightData = rightNode->getData().getDataVal(&size);
            rightData = rightData.substr(0,size);
            
            toDivide = dataStack[0] + rightData;
        }
        
        if(toDivide.size() > lmax)
        {	dataStack.clear();
            dataStack.push_back(toDivide.substr(0,lmin));
            dataStack.push_back(toDivide.substr(lmin,toDivide.size() - lmin));
            
            DiffSkipList update(DiffSkipList::INSERT, outStack[0], verifier, dataStack[1]);
            shortStack.push_back(update);
            DiffSkipList update2(DiffSkipList::MODIFY, precedentNodeStartIndex, verifier, dataStack[0]);
            shortStack.push_back(update2);// 2) otherwise we divide the block into two and send one modify and one insert (always in the reverse order)
        }else{			//1) now there are two options where one is the new string doesnt surpass lmax, then we just modify the precenet block by addition of the new short string.
            DiffSkipList update(DiffSkipList::MODIFY, precedentNodeStartIndex, verifier, toDivide);
            shortStack.push_back(update);
            
        }
        
        dataStack.clear();
        
    }
    //cout<<"dataStack: "<<endl;
    //for(int i = 0 ; i < dataStack.size() ; i++)
    //	cout<<dataStack[i]<<endl;
    //cout<<"--dataStack."<<endl;
    
    //max -> extraSize, min-> modifyCount since first blocks will get modified, other will be inserted or removed
    int extraCount = 0;	int modifyCount = 0;
    if(outStack.size() < dataStack.size()){	extraCount = dataStack.size(); modifyCount = outStack.size();
    }else{ extraCount = outStack.size(); modifyCount = dataStack.size();}
    
    //GENERATE UPDATE STACK FOR THE PARTICULAR DIFF OPERATION
    
    if(outStack.size() < dataStack.size())
    {
        for(int i = dataStack.size() -1; i > modifyCount -1; i--)
        {
            DiffSkipList update(DiffSkipList::INSERT, insertIndex, verifier, dataStack[i]);
            updateStack.push_back(update);
        }
    }else{
        for(int i = 0; i < extraCount - modifyCount; i++)
        {
            DiffSkipList update(DiffSkipList::REMOVE, outStack[extraCount - modifyCount]);
            cout<<"INDEX TO BE REMOVED ::::::::::::::" <<outStack[extraCount - modifyCount]<<endl; //XXX: This may catch a bug later!
            updateStack.push_back(update);
        }
    }
    
    for(int i = modifyCount - 1; i > -1; i--)
    {
        DiffSkipList update(DiffSkipList::MODIFY, outStack[i], verifier, dataStack[i]);
        updateStack.push_back(update);
    }
    
    
    //if the added string is short we need to merge it with the precedent node!
    //We already prepared it above now we need to add this to the update stack which will be sent
    for(int i=0; i<shortStack.size(); i++)
    {
        updateStack.push_back(shortStack[i]);
    }
    //END GENERATE UPDATE STACK FOR THE PARTICULAR DIFF OPERATION
    return updateStack;
}

/*This function is here for later referance*/
vector<DiffSkipList> DiffSkipList::generateDiffsTest(RankBasedSkipList* rbsl, int csi, int ol, string inString, int lmin, int lmax) //unnec
{
    int changeStartIndex = csi;// -1;
    int outLength = ol;// +1;
    vector<DiffSkipList> updateStack;
    vector<DiffSkipList> shortStack;
    int inLength = inString.size();
    Key k(changeStartIndex);
    Node* cN = rbsl->search(&k);
    int remainder = k.getValue() -1;
    /////////////For "short inString" diff operaions to append the new string to the precedent node
    int indexInsidePrecedentNode = changeStartIndex - k.getValue() -1;
    Key key(indexInsidePrecedentNode);
    Node* precedentNode = rbsl->search(&key);
    //////////////
    cout<<"changeStartIndex: "<<changeStartIndex<<endl;
    
    string header = string(cN->getData().getDataVal()).substr(0,k.getValue() ); //was k-1
    
    cout << "header: " << header << endl;
    
    
    int outNodeIndex = changeStartIndex - k.getValue();
    //cout << "outNodeIndex: " << outNodeIndex << endl;
    
    vector<int> outStack;
    outStack.push_back(outNodeIndex);
    
    int outCounted = cN->getData().getLength() - remainder;
    //cout << "outCounted: " << outCounted << endl;
    ///////////////////////////////////////////////////////////////////////////////////// List of node indices
    while(outCounted < outLength)
    {
        //cout << "outNodeIndex: " << outNodeIndex << endl;
        outNodeIndex += cN->getData().getLength();
        //cout << "outNodeIndex: " << outNodeIndex << endl;
        //cout << "cN->getData().getLength(): " << cN->getData().getLength() << endl;
        outStack.push_back(outNodeIndex);
        cN = cN->getLeafAfter();
        //cout << "outCounted: " << outCounted << endl;
        outCounted += cN->getData().getLength();
        //cout << "outCounted: " << outCounted << endl;
        //cout << "cN->getData().getLength(): " << cN->getData().getLength() << endl;
    }
    Node* rightNode = cN->getLeafAfter();
    int insertIndex = outNodeIndex + cN->getData().getLength();
    ///////////////////////////////////////////////////////////////////////////////////// END List of node indices
    long size = -1;
    //cout << "insertIndex: " << insertIndex << endl;
    
    string d = cN->getData().getDataVal(&size);
    //string catS  = string(cat);
    string data = d.substr(0,size);
    
    //cout << "cN.data.size: " << size << endl;
    //cout << "cN.data: " << data << endl;
    
    string footer = data.substr(data.size() - (outCounted - outLength), (outCounted - outLength));
    
    cout << "footer: " << footer << endl;
    
    bool isTooShort = false;
    vector<string> dataStack = DiffSkipList::generateBlocks(header, inString, footer, lmin, lmax, &isTooShort);
    cout<<"isTooShort: " << isTooShort <<endl;
    
    
    if(isTooShort)
    {
        // an index which is inside the precedent node of the node where change started
        // if that index is below 0 that means the change started from the first node, then we need to add the new short block to
        // a node further
        string toDivide;
        int precedentNodeStartIndex = 0;
        if(indexInsidePrecedentNode > 0){
            
            precedentNodeStartIndex = indexInsidePrecedentNode - key.getValue();
            long size = -1;
            string precedentData = precedentNode->getData().getDataVal(&size);
            precedentData = precedentData.substr(0,size);
            //string newDataToBlocks = precedentData + header + inString + footer;
            toDivide = precedentData + dataStack[0];
            
            
            //return updateStack;
        }else{
            //if update occurs at the first block and the inString is short
            //NOT IMPLEMENTED YET!!!
            long size = -1;
            string rightData = rightNode->getData().getDataVal(&size);
            rightData = rightData.substr(0,size);
            
            toDivide = dataStack[0] + rightData;
        }
        
        if(toDivide.size() > lmax)
        {	dataStack.clear();
            dataStack.push_back(toDivide.substr(0,lmin));
            dataStack.push_back(toDivide.substr(lmin,toDivide.size() - lmin));
            
            DiffSkipList update(DiffSkipList::INSERT, outStack[0],  dataStack[1]);
            shortStack.push_back(update);
            DiffSkipList update2(DiffSkipList::MODIFY, precedentNodeStartIndex, dataStack[0]);
            shortStack.push_back(update2);// 2) otherwise we divide the block into two and send one modify and one insert (always in the reverse order)
        }else{			//1) now there are two options where one is the new string doesnt surpass lmax, then we just modify the precenet block by addition of the new short string.
            DiffSkipList update(DiffSkipList::MODIFY, precedentNodeStartIndex, toDivide);
            shortStack.push_back(update);
            
        }
        
        dataStack.clear();
        
    }
    //cout<<"dataStack: "<<endl;
    //for(int i = 0 ; i < dataStack.size() ; i++)
    //	cout<<dataStack[i]<<endl;
    //cout<<"--dataStack."<<endl;
    
    //max -> extraSize, min-> modifyCount since first blocks will get modified, other will be inserted or removed
    int extraCount = 0;	int modifyCount = 0;
    if(outStack.size() < dataStack.size()){	extraCount = dataStack.size(); modifyCount = outStack.size();
    }else{ extraCount = outStack.size(); modifyCount = dataStack.size();}
    
    //GENERATE UPDATE STACK FOR THE PARTICULAR DIFF OPERATION
    
    if(outStack.size() < dataStack.size())
    {
        for(int i = dataStack.size() -1; i > modifyCount -1; i--)
        {
            DiffSkipList update(DiffSkipList::INSERT, insertIndex, dataStack[i]);
            updateStack.push_back(update);
        }
    }else{
        for(int i = 0; i < extraCount - modifyCount; i++)
        {
            DiffSkipList update(DiffSkipList::REMOVE, outStack[extraCount - modifyCount]);
            updateStack.push_back(update);
        }
    }
    
    for(int i = modifyCount - 1; i > -1; i--)
    {
        DiffSkipList update(DiffSkipList::MODIFY, outStack[i], dataStack[i]);
        updateStack.push_back(update);
    }
    
    
    //if the added string is short we need to merge it with the precedent node!
    //We already prepared it above now we need to add this to the update stack which will be sent
    for(int i=0; i<shortStack.size(); i++)
    {
        updateStack.push_back(shortStack[i]);
    }
    //END GENERATE UPDATE STACK FOR THE PARTICULAR DIFF OPERATION
    return updateStack;
}

//generate indices for proof
vector<int> DiffSkipList::indicesToGenerateFirstProofFor(vector<DiffSkipList> diffs)
{
    
    vector<int> indicesToGenerateFirstProofFor;
    
    for(int i = diffs.size()-1 ; i > -1 ; i--)
    {
        if(diffs[i].getType() == DiffSkipList::MODIFY)
            indicesToGenerateFirstProofFor.push_back(diffs[i].getIndex());
        if(diffs[i].getType() == DiffSkipList::INSERT)
        {
            indicesToGenerateFirstProofFor.push_back(diffs[i].getIndex()-1);
            //indicesToGenerateFirstProofFor.push_back(diffs[i]->getIndex());
        }
        if(diffs[i].getType() == DiffSkipList::REMOVE)
        {
            indicesToGenerateFirstProofFor.push_back(diffs[i].getIndex()-1);
            indicesToGenerateFirstProofFor.push_back(diffs[i].getIndex());
        }
    }
    sort(indicesToGenerateFirstProofFor.begin(), indicesToGenerateFirstProofFor.begin()+indicesToGenerateFirstProofFor.size());
    
    return indicesToGenerateFirstProofFor;
    
}

void DiffSkipList::applyDiffsToSL(DPDPFlexProver* prover,RankBasedSkipList* rbsl,vector<DiffSkipList> diffs,string fileName)
{
    char fileNametxt[80] = "";
    strcat(fileNametxt, fileName.c_str());
    
    
    char fileCounter[80]="connectedClients/";
    strcat(fileCounter, prover->getClientIP().c_str());
    strcat(fileCounter, "/counter.txt");
    int blockCounter = Data::readCounterInt(fileCounter/*"serverFolder/counter.txt"*/);
    
    for(int i = 0 ; i < diffs.size() ; i++)
    {
        if(diffs[i].getType() == DiffSkipList::MODIFY){
            char str[100];sprintf(str, "%d", blockCounter);strcat(fileNametxt, str);strcat(fileNametxt, "NEW.txt");
            Data d(diffs[i].getData(),fileNametxt,blockCounter);
            rbsl->modify(Key(diffs[i].getIndex()),d,diffs[i].getTag(),NULL);
            blockCounter++;
        }
        
        if(diffs[i].getType() == DiffSkipList::INSERT){
            char str[100];
            sprintf(str, "%d", blockCounter);
            strcat(fileNametxt, str);
            strcat(fileNametxt, "NEW.txt");
            Data d(diffs[i].getData(),fileNametxt,blockCounter);
            rbsl->insert(Key(diffs[i].getIndex()),d,diffs[i].getTag(),NULL);
            blockCounter++;
        }
        if(diffs[i].getType() == DiffSkipList::REMOVE)
            rbsl->remove(Key(diffs[i].getIndex()));
        
    }
    
    //remove the counter
    ofstream counterFile;
    remove(fileCounter /*"serverFolder/counter.txt"*/);
    counterFile.open(fileCounter/*"serverFolder/counter.txt"*/);
    counterFile << blockCounter;
    counterFile.close();
    
}

void DiffSkipList::applyDiffsToSLforVerifier(RankBasedSkipList* dummyRbsl,vector<DiffSkipList> diffs,string fileName)
{
    int blockCounter = Data::readCounterInt("clientFolder/counter.txt");
    char fileNametxt[80] = "";
    strcat(fileNametxt, fileName.c_str());
    
    
    for(int i = 0 ; i < diffs.size() ; i++)
    {
        
        if(diffs[i].getType() == DiffSkipList::MODIFY){
            char str[100];
            sprintf(str, "%d", blockCounter);
            strcat(fileNametxt, str);
            strcat(fileNametxt, "NEW.txt");
            Data d(diffs[i].getData(),fileNametxt,blockCounter);
            dummyRbsl->modifyWOHashCalculation(Key(diffs[i].getIndex()),d,diffs[i].getTag(),NULL);
            blockCounter++;
        }
        if(diffs[i].getType() == DiffSkipList::INSERT){
            char str[100];
            sprintf(str, "%d", blockCounter);
            strcat(fileNametxt, str);
            strcat(fileNametxt, "NEW.txt");
            Data d(diffs[i].getData(),fileNametxt,blockCounter);
            dummyRbsl->insertWOHashCalculation(Key(diffs[i].getIndex()),d,diffs[i].getTag(),NULL);
            blockCounter++;
        }
        if(diffs[i].getType() == DiffSkipList::REMOVE)
            dummyRbsl->removeWOHashCalculation(Key(diffs[i].getIndex()));
        
    }
    ofstream counterFile;
    remove("clientFolder/counter.txt");
    counterFile.open("clientFolder/counter.txt");
    counterFile << blockCounter;
    counterFile.close();
    
}

/*With this function we aim to divide the string that we will fit in the skip list into blocks with at least min block size and at max the maximum block size*/
vector<string> DiffSkipList::generateBlocks(string header, string str, string footer, int lmin, int lmax, bool* tooShort)
{
    
    
    vector<string> dataStack;
    string inString = header + str + footer;
    string tempStr = inString;
    int tempLength = inString.size();
    
    if(tempLength < lmin)
    {
        cout<<"SHORT"<<endl;
        dataStack.push_back(inString);
        *tooShort = true;
        return dataStack;
    }
    
    while(tempLength > lmax)
    {
        dataStack.push_back(tempStr.substr(0,lmin));
        tempLength -= lmin;
        tempStr = tempStr.substr(lmin, tempStr.size()-lmin);
    }
    dataStack.push_back(tempStr);
    
    *tooShort = false;
    return dataStack;
}

void DiffSkipList::print()
{
    cout <<"TYPE      : " << type << endl;
    cout <<"index : "<< index << endl;
    cout <<"Data size :  " << data.size() << endl;
    if(data.empty())
        cout<<"data		: empty"<<endl;
    else{
        cout<<"data	: " ;
        for(int i = 0; i < data.size(); i++){
            cout<< data[i];
            if(i>9) break; //to deal with big datas. printing first
        }
        cout<<"|...|";
        //to deal with big datas. Printing last
        for(int i = data.size() -10; i < data.size() ; i++){
            cout<< data[i];
        }
        cout<<endl;
        /*
         cout<<"data(last 10)	: " ;
         for(int i = data.size()-11; i < data.size()-1; i++)
         cout<< data[i];
         cout<<endl;
         
         */
    }
    cout <<"Tag      : " << tagOfData << endl;
    
    cout <<" --- "<< endl;
    
}

template <class Archive>
void DiffSkipList::serialize(Archive& ar, const unsigned int ver) {
    ar  	& auto_nvp(type)
    & auto_nvp(index)
    & auto_nvp(data)
    & auto_nvp(tagOfData)
    ;
}


int DiffSkipList::getIndex() {return index;}
vector<char> DiffSkipList::getData() {return data;}
ZZ DiffSkipList::getTag() { return tagOfData; }
Data* DiffSkipList::getDatap() { return d; }


#endif /*_DiffSkipList_H_*/
