#ifndef _PacketClient_H_
#define _PacketClient_H_

#include "PacketBase.h"
#include "ProofNode.h"

class PacketClient : public PacketBase {
	
/*------------------------------------------------------------------------
 * Program:   	PacketClient
 *
 * Purpose:   	Packet that is created by the client and is sent to the server
 *
 *
 * Note:      	Some fields(*) are for basic test cases which will be removed later
 *
 *------------------------------------------------------------------------
 */
	
public:

	PacketClient( TYPE t)
	: PacketBase(t)  {}
	PacketClient() {}
    PacketClient( TYPE t, int c, int i, int s,int l, ZZ n, ZZ gg, char* f, vector<char> b, vector<ZZ> ts, vector<DiffSkipList> aDiffs, int id = 0); //constructor for INIT PACKET
    PacketClient( TYPE t, int c, char* f, ZZ n, ZZ gg, int id = 0); //constructor for INIT PACKET
    PacketClient( TYPE t, vector<char> b, vector<ZZ> ts, int id = 0); //constructor for TRANSMISSION PACKET
    PacketClient( TYPE t, int c, int rndi, int id = 0); //constructor for CHALLENGE / REMOVE PACKET
    PacketClient( TYPE t, int rnd, vector<DiffSkipList> aDiffs, int id = 0/*,vector<ZZ> tags*/); //constructor for COMMIT PACKET
    ~PacketClient();
    void print();

    int getLevel();
    int getCount();
    int getSize();
    int getSeed();
    int getClientID();
    void setClientID(int id);
    ZZ getN();
    ZZ getTag();
    ZZ getG();
    string getFileName();
    int getIndex();
    vector<ZZ> getTags();
    vector<char> getBlock();
    void setBlock(vector<char> b);
    vector<DiffSkipList> getAllDiffs();
    
private:
	
	int count; //*
	int index; //*
	int clientID;
	int seed;
	int level;
	ZZ N_Tag; 			
	ZZ g; 	
	string fileName;		
	vector<char> block;		// File transmission		
	vector<ZZ> Tags;		//used both in fileTransfer and diffupdatepackages
	vector<DiffSkipList> allDiffs; //The diff operations ready to be applied and to get proofs generated

	friend class boost::serialization::access;
	template <class Archive> 
    void serialize(Archive& ar, const unsigned int ver);

};


PacketClient::PacketClient( TYPE t, int c, int i, int s,int l, ZZ n, ZZ gg, char* f, vector<char> b, vector<ZZ> ts, vector<DiffSkipList> aDiffs, int id) //constructor for INIT PACKET
: PacketBase(t),count(c),index(i),seed(s),level(l),fileName(f),N_Tag(n),g(gg),block(b),Tags(ts),allDiffs(aDiffs),clientID(id){}

PacketClient::PacketClient( TYPE t, int c, char* f, ZZ n, ZZ gg, int id) //constructor for INIT PACKET
: PacketBase(t),count(c),seed(NULL),fileName(f),N_Tag(n),g(gg),block(NULL),index(NULL),Tags(NULL),allDiffs(NULL),clientID(id){}

PacketClient::PacketClient( TYPE t, vector<char> b, vector<ZZ> ts, int id) //constructor for TRANSMISSION PACKET
: PacketBase(t),count(NULL),seed(NULL),fileName(""),N_Tag(to_ZZ(1)),g(to_ZZ(1)),block(b),index(NULL),Tags(ts),allDiffs(NULL),clientID(id){}


PacketClient::PacketClient( TYPE t, int c, int rndi, int id) //constructor for CHALLENGE / REMOVE PACKET
:PacketBase(t),count(c),fileName(""),N_Tag(to_ZZ(1)),g(to_ZZ(1)),block(NULL),index(NULL),Tags(NULL),allDiffs(NULL),clientID(id) {
    if(t == CHALLENGE) seed = rndi;
    if(t == REMOVE) index = rndi;
}

PacketClient::PacketClient( TYPE t, int rnd, vector<DiffSkipList> aDiffs, int id/*,vector<ZZ> tags*/) //constructor for COMMIT PACKET
: PacketBase(t),count(NULL),seed(rnd),fileName(""),N_Tag(to_ZZ(1)),g(to_ZZ(1)),block(NULL),index(NULL),Tags(NULL),allDiffs(aDiffs),clientID(id){}

PacketClient::~PacketClient() {}

void PacketClient::print() {
    cout<<"<-----------2Server------------>"<<endl;
    PacketBase::print();
    cout<<"count        : "<<count		<<endl;
    cout<<"index        : "<<index		<<endl;
    cout<<"seed         : "<<seed		<<endl;
    cout<<"N|Tag        : "<<N_Tag 		<<endl;
    cout<<"fileName     : "<<fileName	<<endl;
    cout<<"g            : "<<g			<<endl;
    if(Tags.empty())
        cout<<"Tags		: empty"<<endl;
    else cout<<"Tags.size	: " << Tags.size() <<endl;
    
    if(block.empty())
        cout<<"block		: empty"<<endl;
    else{
        cout<<"block(first 10)	: " ;
        for(int i = 0; i < 10; i++)
            cout<< block[i];
        cout<<endl;
        cout<<"block(last 10)	: " ;
        for(int i = block.size()-11; i < block.size()-1; i++)
            cout<< block[i];
        cout<<endl;
				}
				
    if(Tags.empty())
        cout<<"Tasgs		: empty"<<endl;
    else{
        cout<<"Tags size    : "<< Tags.size()<<endl;
        cout<<"Tags(first)	: " ;
        cout<< Tags[0]<<endl;
        cout<<"Tags(last)	: " ;
        cout<< Tags[Tags.size()-1]<<endl;
				}
				
    cout<<"<------------------------------>"<<endl;
				
}

int PacketClient::getLevel() { return level; }

int PacketClient::getCount() { return count; }

int PacketClient::getSize() { return count; }

int PacketClient::getSeed() { return seed; }

int PacketClient::getClientID() { return clientID; }

void PacketClient::setClientID(int id) { clientID = id; }

ZZ PacketClient::getN() { return N_Tag; }

ZZ PacketClient::getTag() { return N_Tag; }

ZZ PacketClient::getG() { return g; }

string PacketClient::getFileName() { return fileName; }

int PacketClient::getIndex() { return index; }

vector<ZZ> PacketClient::getTags() { return Tags; }

vector<char> PacketClient::getBlock() { return block; }

void PacketClient::setBlock(vector<char> b) { block = b; }

vector<DiffSkipList> PacketClient::getAllDiffs() { return allDiffs; }

template <class Archive>
void PacketClient::serialize(Archive& ar, const unsigned int ver) {
    PacketBase::serialize(ar,ver);
    ar
    & auto_nvp(count)
    & auto_nvp(index)
    & auto_nvp(clientID)
    & auto_nvp(seed)
    & auto_nvp(level)
    & auto_nvp(N_Tag)
    & auto_nvp(g)
    & auto_nvp(fileName)
    & auto_nvp(block)
    & auto_nvp(Tags)
    & auto_nvp(allDiffs)
    ;
}


#endif /*_PacketClient_H_*/
