#ifndef _PacketServer_H_
#define _PacketServer_H_

#include "PacketBase.h"
/*------------------------------------------------------------------------
 * Program:   	PacketServer
 *
 * Purpose:   	Packet that is created by the Server and is sent to the client
 *
 *
 * Note:      	Some fields(*) are for basic test cases which will be removed later
 *
 *------------------------------------------------------------------------
 */


class PacketServer : public PacketBase {
	
public:

    PacketServer(TYPE t);
    PacketServer();
    PacketServer( TYPE t, int s, ZZ z, vector<ProofNode> pv, vector<ProofNode> pvu,	vector<ZZ> ts,	vector<ZZ> TU,	vector<char> b, vector<CommitProof> pr);
    PacketServer( TYPE t, int s); //constructor for INFO
    PacketServer( TYPE t, int c, vector<ProofNode> pv,	vector<ZZ> ts, ZZ blockSum); //constructor for PROOF PACKET
    PacketServer( TYPE t, vector<ProofNode> pv, vector<ProofNode> pvu, vector<ZZ> ts,	vector<ZZ> TU, ZZ blockSum); //constructor for UPDATEPROOF PACKET
    PacketServer( TYPE t, vector<ProofNode> pv, vector<ZZ> ts, ZZ blockSum, ZZ newHashOfRoot); //constructor for dummySkipList UPDATEPROOF PACKET
    PacketServer( TYPE t, ZZ z); //constructor for REWRITEPROOF PACKET
    PacketServer( TYPE t, vector<CommitProof> pr); //constructor for REWRITEPROOF PACKET
    ~PacketServer();
    
    void print();
	
    int getCount();
    int getSeed();
    ZZ getRewriteProof();
    vector<ZZ> getTags();
    ZZ getBlockSum();
    ZZ getUpdatedHash();
    vector<char> getBlock();
    vector<ZZ> getTagsUpdated();
    vector<ProofNode> getProofVec();
    vector<ProofNode> getProofVecUpdated();
    vector<CommitProof> getCommitProofs();
    void setNewHash(ZZ nHash);
	
private:

	int seed; 
	ZZ zzValue;
	ZZ newHash;	
	vector<ProofNode> proofVector; 			//*
	vector<ProofNode> proofVectorUpdated; 	//*
	vector<ZZ> Tags;						//*
	vector<ZZ> TagsUpdated;					//*
	vector<char> block;		// For file download
	vector<CommitProof> proofs;	//This field includes all proofs possible for a sequence of updates
	
	friend class boost::serialization::access;
	template <class Archive> 
    void serialize(Archive& ar, const unsigned int ver);

};

//Method and function definitions//

PacketServer::PacketServer(TYPE t)
: PacketBase(t)  {}

PacketServer::PacketServer() {}

PacketServer::PacketServer( TYPE t, int s, ZZ z, vector<ProofNode> pv, vector<ProofNode> pvu,	vector<ZZ> ts,	vector<ZZ> TU,	vector<char> b, vector<CommitProof> pr)
: PacketBase(t),seed(s),zzValue(z),proofVector(pv), proofVectorUpdated(pvu),Tags(ts),TagsUpdated(TU),block(b),proofs(pr),newHash(NULL){}

PacketServer::PacketServer( TYPE t, int s) //constructor for INFO
: PacketBase(t),seed(s),zzValue(to_ZZ(1)),proofVector(NULL), proofVectorUpdated(NULL),Tags(NULL),TagsUpdated(NULL),block(NULL),proofs(NULL),newHash(NULL){}

PacketServer::PacketServer( TYPE t, int c, vector<ProofNode> pv,	vector<ZZ> ts, ZZ blockSum) //constructor for PROOF PACKET
: PacketBase(t),seed(c),zzValue(blockSum), proofVectorUpdated(NULL),TagsUpdated(NULL),block(NULL),proofs(NULL),newHash(NULL){
    for(int i= 0; i< ts.size(); i++)
        Tags.push_back(ts[i]);
    Tags.resize(Tags.size());
    for(int i= 0; i< pv.size(); i++)
        proofVector.push_back(pv[i]);
    
    //cout<<"size: : : " << proofVector.size()<<endl;
    //proofVector.resize(20/*proofVector.size()*/);
    
}

PacketServer::PacketServer( TYPE t, vector<ProofNode> pv, vector<ProofNode> pvu, vector<ZZ> ts,	vector<ZZ> TU, ZZ blockSum) //constructor for UPDATEPROOF PACKET
: PacketBase(t),seed(NULL),zzValue(blockSum),proofVector(pv), proofVectorUpdated(pvu),Tags(ts),TagsUpdated(TU),block(NULL),proofs(NULL),newHash(NULL){}

PacketServer::PacketServer( TYPE t, vector<ProofNode> pv, vector<ZZ> ts, ZZ blockSum, ZZ newHashOfRoot) //constructor for dummySkipList UPDATEPROOF PACKET
: PacketBase(t),seed(NULL),zzValue(blockSum),proofVector(pv), proofVectorUpdated(NULL),Tags(ts),TagsUpdated(NULL),block(NULL),proofs(NULL),newHash(newHashOfRoot) /*hashValue(newHashOfRoot)*/{}

PacketServer::PacketServer( TYPE t, ZZ z) //constructor for REWRITEPROOF PACKET
: PacketBase(t),seed(NULL),zzValue(z),proofVector(NULL), proofVectorUpdated(NULL),Tags(NULL),TagsUpdated(NULL),block(NULL),proofs(NULL),newHash(NULL){}

PacketServer::PacketServer( TYPE t, vector<CommitProof> pr) //constructor for REWRITEPROOF PACKET
: PacketBase(t),seed(NULL),zzValue(to_ZZ(1)),proofVector(NULL), proofVectorUpdated(NULL),Tags(NULL),TagsUpdated(NULL),block(NULL),proofs(pr),newHash(NULL){}


PacketServer::~PacketServer() {}

void PacketServer::print() {
    cout<<"<-----------2Client------------>"<<endl;
    PacketBase::print();
    cout<<"Seed         : "<<seed		<<endl;
    //cout<<"zzValue      : "<<zzValue	 		<<endl;  //As blockSum is huge(ie 16kb), lets dont print it
    if(block.empty())
        cout<<"block		: empty"<<endl;
    else{
        cout<<"block(first 10)	: " ;
        for(int i = 0; i < 10; i++)
            cout<< block[i];
        cout<<endl;
        cout<<"block(last 10)	: " ;
        for(int i = block.size()-11; i < block.size()-1; i++)
            cout<< block[i];
        cout<<endl;
				}
				
    if(Tags.empty())
        cout<<"Tasgs		: empty"<<endl;
    else{
        cout<<"Tags size	:"<< Tags.size()<<endl;
        cout<<"Tags(first)	: " ;
        cout<< Tags[0]<<endl;
        cout<<"Tags(last)	: " ;
        cout<< Tags[Tags.size()-1]<<endl;
				}
				
    cout<<"<------------------------------>"<<endl;
}

int PacketServer::getCount() { return seed; }
int PacketServer::getSeed() { return seed; }
ZZ PacketServer::getRewriteProof() { return zzValue; }
vector<ZZ> PacketServer::getTags() { return Tags; }
ZZ PacketServer::getBlockSum() { return zzValue; }
ZZ PacketServer::getUpdatedHash() { return newHash; }
vector<char> PacketServer::getBlock() { return block; }
vector<ZZ> PacketServer::getTagsUpdated() { return TagsUpdated; }
vector<ProofNode> PacketServer::getProofVec() {
    return proofVector;
}
vector<ProofNode> PacketServer::getProofVecUpdated() { return proofVectorUpdated; }
vector<CommitProof> PacketServer::getCommitProofs() { return proofs; }
void PacketServer::setNewHash(ZZ nHash) { newHash = nHash; }

template <class Archive>
void PacketServer::serialize(Archive& ar, const unsigned int ver) {
    PacketBase::serialize(ar,ver);
    ar
    & auto_nvp(seed)
    & auto_nvp(zzValue)
    & auto_nvp(newHash)
    & auto_nvp(proofVector)
    & auto_nvp(proofVectorUpdated)
    & auto_nvp(Tags)
    & auto_nvp(TagsUpdated)
    & auto_nvp(block)
    & auto_nvp(proofs)
    ;
}

#endif /*_PacketClient_H_*/
