#ifndef _ServerDPDPmain_H_
#define _ServerDPDPmain_H_

using namespace std;
#include "DPDPFlexProver.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <map>
#include <boost/asio.hpp>
#include <boost/bind.hpp>
#include <boost/lexical_cast.hpp>
#include <iostream>
#include <vector>
#include "connection.hpp" // Must come before boost/serialization headers.
#include <boost/serialization/vector.hpp>
#include <boost/thread.hpp>
#include <math.h>

#include "GroupRSA.h"
#include "RankBasedSkipList.h"

#include "PacketClient.h"
#include "PacketServer.h"
#include <CommitProof.h>



//#define _DEFINE_DEPRECATED_HASH_CLASSES 0
#define PROTOPORT       5195           /* default protocol port number */
#define QLEN            6               /* size of request queue        */
//using namespace boost;
using boost::asio::ip::tcp;
/*------------------------------------------------------------------------
 * Program:   	ServerDPDPmain
 *
 * Purpose:   	waits for an initialization for a client, when initialization command arrives initialize FlexDPDP variables,
 * 				lets Client use the menu where she can send her file, challenge her file and commit changes done
 *
 *
 * Note:      	This is a basic server class where the server acts according to a packet arrived
 *				This code is not meant to be secure and mostly implemented for test purposes
 *
 // Partially from http://netbook.cs.purdue.edu
 *------------------------------------------------------------------------
 */
// Partially:
// Copyright (c) 2003-2012 Christopher M. Kohlhoff (chris at kohlhoff dot com)
//
// Distributed under the Boost Software License, Version 1.0. (See accompanying
// file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)

namespace s11n_example {
    
    class ServerDPDPmain {
        
    public:
        
        ServerDPDPmain(boost::asio::io_service& io_service, unsigned short port, int stt = 0, int* c = NULL, RankBasedSkipList* r = NULL);
        
        /// Handle completion of a accept operation.
        void handle_accept(const boost::system::error_code& e, connection_ptr conn, int* c = NULL);
        
        /// Handle completion of a read operation.
        void handle_read(const boost::system::error_code& e, connection_ptr con, int* c = NULL/*, DPDPFlexProver* prover*/);
        /// Handle completion of a write operation.
        void handle_write(const boost::system::error_code& e, connection_ptr conn, int* c);
        
    private: 
        
        /*PL test skip list*/
        RankBasedSkipList *rbsl;
        /*DPDP RELATED FIELDS*/
        ZZ N,g;
        int seed;
        int size;
        //hash_map<int, ClientState> state_db;		
        int serverTimeTest;
        char* directory;
        string fileNameString;
        char* fileName;
        string fileNameStringBase;
        char* fileNameBase;
        // The acceptor object used to accept incoming socket connections.
        boost::asio::ip::tcp::acceptor acceptor_;
        static DPDPFlexProver prover;
        PacketClient packetFromClient;
        map<string,DPDPFlexProver> connectedClients;
        
    };
    
    
    DPDPFlexProver ServerDPDPmain::prover;
    
    //Method and function definitions//
    
    ServerDPDPmain::ServerDPDPmain(boost::asio::io_service& io_service, unsigned short port, int stt, int* c, RankBasedSkipList* r)
    : acceptor_(io_service,
    boost::asio::ip::tcp::endpoint(boost::asio::ip::tcp::v4(), port))
    {
        rbsl = r;
        serverTimeTest = stt;
        
        // Start an accept operation for a new connection.
        connection_ptr new_conn(new connection(acceptor_.get_io_service()));
        acceptor_.async_accept(new_conn->socket(),
                               boost::bind(&ServerDPDPmain::handle_accept, this,
                                           boost::asio::placeholders::error, new_conn, c));
    }
    
    /// Handle completion of a accept operation.
    void ServerDPDPmain::handle_accept(const boost::system::error_code& e, connection_ptr conn, int* c)
    {
        if (!e)
        {
            
            //	cout<<"accept! :async_read: "<<endl;
            conn->async_read(packetFromClient,
                             boost::bind(&ServerDPDPmain::handle_read, this,
                                         boost::asio::placeholders::error, conn, c/*, &prover*/));
            //	cout<<"async_read passed. "<<endl;
            
        }
        
        // Start an accept operation for a new connection.
        connection_ptr new_conn(new connection(acceptor_.get_io_service()));
        acceptor_.async_accept(new_conn->socket(),
                               boost::bind(&ServerDPDPmain::handle_accept, this,
                                           boost::asio::placeholders::error, new_conn, c));
    }
    
    /// Handle completion of a read operation.
    void ServerDPDPmain::handle_read(const boost::system::error_code& e, connection_ptr con, int* c/*, DPDPFlexProver* prover*/)
    {
        if (!e)
        {
            int dummy = 0;
            char temp[100];
            string clientIP;
            const char* ip = con->socket().remote_endpoint().address().to_string().c_str(); //XXX: id ekte
            sprintf(temp,"%s%d",ip,packetFromClient.getClientID());
            //clientIP = string(temp);
            clientIP = string( temp ) ;
            
            //cout<<"client identifier: "<<clientIP<<endl;
            //string clientIP = boost::lexical_cast<std::string>(con->socket().remote_endpoint());XXX OZAN_clientIP with PORT
            
            if(packetFromClient.getType() == PacketClient::INIT)
            {
                if(connectedClients.find(clientIP)!=connectedClients.end())
                {
                    //client is already connected no need to create folders
                    cout<<"client found in the list of hashmap"<<endl;
                    PacketServer p( PacketServer::INFO, -1);
                    con->async_write(p,
                                     boost::bind(&ServerDPDPmain::handle_write, this,
                                                 boost::asio::placeholders::error, con, &dummy));
                    
                }else{
                    //////////////////////////////////////////////////just setting file name...
                    
                    //char fname[80] = "serverFolder/";
                    char folderName[80] = "connectedClients/";
                    char folderNameBase[80] = "connectedClients/";
                    char folderNameBasetxt[80] = "connectedClients/";
                    //char fnametxt[84] = "serverFolder/";
                    char mkdir[80] = "mkdir ";
                    //strcat(fname, packetFromClient.getFileName().c_str());
                    strcat(folderName, clientIP.c_str());
                    strcat(folderName, "/");
                    strcat(folderName, packetFromClient.getFileName().c_str());
                    //strcat(fnametxt, packetFromClient.getFileName().c_str());
                    strcat(folderNameBasetxt, clientIP.c_str());
                    strcat(folderNameBasetxt, "/");
                    strcat(folderNameBasetxt, packetFromClient.getFileName().c_str());
                    strcat(folderNameBasetxt, ".txt");
                    
                    strcat(folderNameBase, clientIP.c_str());
                    
                    strcat(mkdir, folderNameBase);
                    
                    system(mkdir);
                    
                    fileNameString = string(folderNameBasetxt);
                    fileNameStringBase = string(folderName);
                    cout << "fileNameStringBase*: " << fileNameStringBase << endl;
                    cout << "fileNameString*: " << fileNameString << endl;
                    ///////////////////////////////////////////////////////////
                    
                    packetFromClient.print();
                    cout<<"INITIALIZING STORAGE"<<endl;
                    size = packetFromClient.getSize();
                    
                    DPDPFlexProver proverT(&seed, packetFromClient.getN(), packetFromClient.getG());
                    proverT.setFileName(fileNameString);
                    proverT.setFileNameBase(fileNameStringBase);
                    proverT.setClientIP(clientIP);
                    
                    proverT.setNumBlocks(size);
                    
                    connectedClients[clientIP]=proverT;
                    
                    
                    PacketServer p( PacketServer::INFO, seed);
                    con->async_write(p,
                                     boost::bind(&ServerDPDPmain::handle_write, this,
                                                 boost::asio::placeholders::error, con, &dummy));
                    
                }
                
            }
            if(packetFromClient.getType() == PacketClient::REINIT)
            {
                packetFromClient.print();
                /*This may be used in case of connection loss.*/
            }
            if(packetFromClient.getType() == PacketClient::TRANSMISSION)
            {
                
                if(connectedClients.find(clientIP)==connectedClients.end())
                {
                    //client is not in the list and try to transmit a packet from init
                    cout<<"Client is not in the list of hashmap"<<endl;
                    
                }else{
                    
                    DPDPFlexProver* prover = &connectedClients[clientIP];
                    //cout << "fileNameStringBase*: " << fileNameStringBase << endl;
                    //cout << "fileNameString*: " << fileNameString << endl;
                    
                    packetFromClient.print();
                    cout<<"TRANSMISSION STARTED"<<endl;
                    
                    //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
                    //TAKE FILE FROM PACKET TURN IT TO BLOCKS BEFORE REWRITE <vector<char> to file>
                    //I added this feature to class Data for further use!
                    //Then devide it into blocks and send to full rewrite
                    //instead of fileName!!     <ozan>	http://www.cplusplus.com/forum/beginner/76436/
                    
                    char fnameTemp[80] = "";
                    //strcat(fnameTemp, fileNameString.c_str());
                    strcat(fnameTemp, prover->getFileName().c_str());
                    Data d(packetFromClient.getBlock(), fnameTemp/*"serverFolder/aaaBlockNotDividedYet.txt"*/);
                    
                    char fnamebaseTemp[80] = "";
                    //strcat(fnamebaseTemp, fileNameStringBase.c_str());
                    strcat(fnamebaseTemp, prover->getFileNameBase().c_str());
                    
                    
                    ZZ rootHash =prover->divideAndBuildSkipList_FullRewritePL(seed, /*"serverFolder/aaaBlockNotDividedYet"*/fnamebaseTemp, packetFromClient.getTags() ,DPDPCONST::defaultBlockLength, rbsl );
                    //ZZ rootHash =prover->divideAndBuildSkipList_FullRewrite(seed, /*"serverFolder/aaaBlockNotDividedYet"*/fnamebaseTemp, packetFromClient.getTags() ,DPDPCONST::defaultBlockLength );
                    
                    
                    
                    Node* rbn = prover->getSL()->search(1);
                    
                    
                    PacketServer p( PacketServer::REWRITEPROOF, rootHash);
                    p.print();cout<<"TRANSMISSION PROOF SENT"<<endl;
                    
                    con->async_write(p,
                                     boost::bind(&ServerDPDPmain::handle_write, this,
                                                 boost::asio::placeholders::error, con, &dummy));
                }
                
            }
            
            /*Generating proof for a challenge seed*/
            if(packetFromClient.getType() == PacketClient::CHALLENGE)
            {
                
                if(connectedClients.find(clientIP)==connectedClients.end())
                {
                    //client is not in the list and try to transmit a packet from init
                    cout<<"Client is not in the list of hashmap-packet discarded"<<endl;
                    
                }else{
                    //startTimer();
                    DPDPFlexProver* prover = &connectedClients[clientIP];
                    //packetFromClient.print();
                    //cout<<"CHALLENGE ARRIVED"<<endl;
                    
                    int chalCount = DPDPCONST::numOfChallenges;
                    int defaultBlockLength = DPDPCONST::defaultBlockLength;
                    vector<int> challenges;
                    int chal;
                    /*GENERATE CHALLENGES*/
                    srand( packetFromClient.getSeed() );
                    
                    challenges.clear();
                    
                    while(challenges.size() < chalCount) {  //for flexlist tries
                        chal = (rand() % ((prover->getNumBlocks() - 1) * defaultBlockLength));
                        chal = chal - chal % defaultBlockLength;
                        vector<int>::iterator it = std::find(challenges.begin(), challenges.end(), chal);
                        if (it != challenges.end() || chal == 0)
                            continue;
                        challenges.push_back(chal);
                    }
                    //cout<<"CHALLENGED INDICES GENERATED"<<endl;
                    
                    sort(challenges.begin(), challenges.begin()+chalCount);
                    //cout<<"Packet::CHALLENGEs generated by using seed: "<< packetFromClient.getSeed() <<endl;
                    vector<ProofNode> proofVector;
                    vector<ZZ> chalBlocks;
                    vector<ZZ> chalTags;
                    ZZ chalBlockSum;
                    //cout<<"Start GENERATE PROOF AND BLOCKSUM"<<endl;
                    /*GENERATE PROOF AND BLOCKSUM*/
                    //					cout<<endl;
                    //					cout<<"MULTIPROOF IS BEING GENERATED"<<endl;
                    char fnamebaseTemp[80] = "";
                    strcat(fnamebaseTemp, prover->getFileNameBase().c_str());
                    proofVector = prover->
                    
                    //For test on PlanetLab
                    GenMultiProofPL(challenges, &chalTags, &chalBlocks, /*"serverFolder/aaaBlockNotDividedYet"*/fnamebaseTemp);
                    //GenMultiProof(challenges, &chalTags, &chalBlocks, /*"serverFolder/aaaBlockNotDividedYet"*/fnamebaseTemp);
                    
                    usleep(550); //SIMULATION time spent for one update *simulation of  a 20KB update... For fast test results this simulation was necessary
                    
                    
                    SetSeed( packetFromClient.getSeed() );
                    //					cout<<"SEEEEEEEEEEEEED "<< packetFromClient.getSeed() <<endl;
                    chalBlockSum = prover->genBlockSum(packetFromClient.getSeed(), chalBlocks, DPDPCONST::stat);
                    //cout<<"END GENERATE PROOF AND BLOCKSUM"<<endl;
                    
                    //int numberOfProofNodesPerPacket = 15;
                    //int proofNodesNeeded = ceil ( proofVector.size() / numberOfProofNodesPerPacket );
                    //vector<ProofNode> tempProofVector;
                    
                    //cout<<"Packet::PROOF "<<proofNodesNeeded<<" ready to be sent!"<<endl;
                    PacketServer p( PacketServer::PROOF, 0/*proofNodesNeeded--*/, proofVector, chalTags, chalBlockSum );
                    p.setNewHash(rbsl->getRoot()->getHash()); //SIMULATION
                    
                    /*if(serverTimeTest)
                     {
                     float fTime = printTimer(0, "ans:challenge=bs+pg");
                     //cout<<fTime<<endl;
                     ofstream fTimeFile;
                     char str[100];
                     sprintf(str, "serverFolder/serveTime%d.txt", serverTimeTest);
                     fTimeFile.open(str, std::fstream::app);
                     fTimeFile << fTime <<endl;
                     fTimeFile.close();
                     }*/
                    //							p.print();
                    //cout<<"MULTIPROOF SENT"<<endl;
                    
                    
                    
                    
                    /*HEMEN CEVABI YOLLAMADAN ÖNCE*/
                    
                    
                    con->async_write(p,
                                     boost::bind(&ServerDPDPmain::handle_write, this,
                                                 boost::asio::placeholders::error, con, c));
                }
            }
            
            /*Basic test case for the client to send a single modify, this piece of code is an old version, but left here for possible later referance*/
            if(packetFromClient.getType() == PacketClient::MODIFY)
            {
                if(packetFromClient.getCount() == 1)
                {
                    packetFromClient.print();
                    //						cout<<"MODIFICATION REQUEST ARRIVED"<<endl;
                    ZZ tag = packetFromClient.getTag();
                    int modIndex = packetFromClient.getIndex();
                    
                    int blockCounter=0;//XXX:will be read from file and be written to file at the end XXX
                    blockCounter = Data::readCounterInt("serverFolder/counter.txt") + 1;
                    
                    char isim[80] = "";
                    char str[100];
                    sprintf(str, "%d", blockCounter);
                    
                    char fnamebaseTemp[80] = "";
                    strcat(fnamebaseTemp, fileNameStringBase.c_str());
                    strcat(isim, /*"serverFolder/aaaBlockNotDividedYet"*/fnamebaseTemp);
                    strcat(isim, str);
                    strcat(isim, ".txt");// = fileName + blockCounter + ".txt";
                    
                    vector<char> block = packetFromClient.getBlock();
                    Data d(block, isim, blockCounter);
                    //Delete counter file and recreate with new counter...
                    //		if( remove( "serverFolder/counter.txt" ) != 0 )
                    //			cout<< "Error deleting file"<<endl;
                    
                    ofstream counterFile;
                    counterFile.open("serverFolder/counter.txt");
                    counterFile << blockCounter;
                    counterFile.close();
                    vector<ProofNode> p;
                    vector<ProofNode> pupv;
                    vector<ZZ> tsB;
                    vector<ZZ> ts;
                    vector<ZZ> chalBlocks;
                    vector<int> challenges;
                    challenges.push_back(modIndex);
                    ZZ chalBlockSum;
                    
                    //char fnamebaseTemp[80] = "";
                    //strcat(fnamebaseTemp, fileNameStringBase.c_str());
                    p = ServerDPDPmain::prover.GenMultiProof(challenges, &tsB, &chalBlocks,/*"serverFolder/aaaBlockNotDividedYet"*/fnamebaseTemp);
                    
                    chalBlockSum = ServerDPDPmain::prover.genBlockSum(seed, chalBlocks, DPDPCONST::stat);
                    ServerDPDPmain::prover.updateModify(modIndex, 0, tag, d);
                    chalBlocks.clear();
                    
                    
                    pupv = ServerDPDPmain::prover.GenMultiProof(challenges, &ts, &chalBlocks, /*"serverFolder/aaaBlockNotDividedYet"*/fnamebaseTemp);
                    //chalBlockSum = ServerDPDPmain::prover.genBlockSum(pack.getSeed(), chalBlocks, DPDPCONST::stat);
                    
                    
                    cout<<"MODIFICATION REQUEST ARRIVED"<<endl;
                    
                    PacketServer pa( PacketServer::UPDATEPROOF, p, pupv, tsB, ts, chalBlockSum);
                    
                    
                    //cout<<"Packet::PROOF ready to be sent!"<<endl;
                    pa.print();
                    //						cout<<"PROOF SENT"<<endl;
                    con->async_write(pa,
                                     boost::bind(&ServerDPDPmain::handle_write, this,
                                                 boost::asio::placeholders::error, con, &dummy));
                }else{
                    //add to prover vector so that when count is 1 we can update all and send proof back
                    //FOR NOW WE JUST SEND PROOF FOR ONE UPDATE  ^ı
                }
            }
            
            /*Same case with modify*/
            if(packetFromClient.getType() == PacketClient::INSERT)
            {
                //For one by one update tests
            }
            
            /*DiffSkipList class is used to generate necessary proofs for the alterations done at the server side with the request of the client*/
            if(packetFromClient.getType() == PacketClient::COMMIT)
            {
                
                DPDPFlexProver* prover = &connectedClients[clientIP];
                
                vector<DiffSkipList> ad = packetFromClient.getAllDiffs();
                
                vector<ZZ> tags;
                vector<ZZ> fileBlocks;
                //bloc
                vector<ProofNode> firstProof = prover->GenMultiProof(DiffSkipList::indicesToGenerateFirstProofFor(ad), &tags, &fileBlocks);
                
                ZZ chalBlockSum = prover->genBlockSum(packetFromClient.getSeed(), fileBlocks, DPDPCONST::stat);
                
                SetSeed( packetFromClient.getSeed() );
                srand( to_int( packetFromClient.getSeed() ) );
                
                DiffSkipList::applyDiffsToSL(prover,prover->getRBSL(),ad,prover->getFileNameBase());

                prover->getRBSL()->printRB();
                
                PacketServer pa( PacketServer::COMMITPROOF, firstProof,tags,chalBlockSum,prover->getRBSL()->getRoot()->getHash());
                con->async_write(pa,
                                 boost::bind(&ServerDPDPmain::handle_write, this,
                                             boost::asio::placeholders::error, con, &dummy));
            }
            
            
        }
        else
        {
            // An error occurred.
            cout<<"?"<<endl;
            std::cerr << e.message() << std::endl;
        }
        
        // Since we are not starting a new operation the io_service will run out of
        // work to do and the client will exit.
    }
    
    /// Handle completion of a write operation.
    void ServerDPDPmain::handle_write(const boost::system::error_code& e, connection_ptr conn, int* c)
    {
        /*The method which is called after each and every response*/
        if(c != NULL)
        {
            *c = *c + 1;
            
            if(*c % 100 == 0){
                ofstream fTimeFile;
                char str[100];
                sprintf(str, "serverFolder/timeFor100s%d.txt", serverTimeTest);
                fTimeFile.open(str, std::fstream::app);
                
                fTimeFile << *c << " "<< boost::posix_time::second_clock::local_time() <<endl;
                fTimeFile.close();
            }
            
        }
        
        // Nothing to do. The socket will be closed automatically when the last
        // reference to the connection object goes away.
    }

    
} // namespace s11n_example
#endif /*_ServerDPDPmain_H_*/
